from os import path
import os
import numpy as np
from multiprocessing import Pool


def populate_im(parameters):
    from .IO import imread, SpatialImage

    path, pos = parameters
    return imread(path)[:, :, 0]


def read_folder(folder, parallel):
    from .IO import imread, SpatialImage

    data_files = sorted(os.listdir(folder))
    nb_images = len(data_files)
    first_image = imread(path.join(folder, data_files[0]))
    to_read = []
    for i, p in enumerate(data_files):
        to_read.append([path.join(folder, p), i])  # (i, path.join(folder, p)))

    if parallel:
        pool = Pool()
        out = pool.map(populate_im, to_read)
        pool.close()
        pool.terminate()
    else:
        out = []
        for param in to_read:
            out.append(populate_im(param))
    return SpatialImage(
        np.array(out).transpose(1, 2, 0), voxelsize=first_image.voxelsize
    )



from pyklb import readfull, writefull, readheader
from .spatial_image import SpatialImage
import numpy as np


def read_klb(filename, SP_im=True):
    tmp = readfull(filename)
    if len(tmp.shape) == 2:
        tmp = tmp.reshape((1,) + tmp.shape)
    if SP_im:
        im = SpatialImage(tmp.transpose(2, 1, 0), copy=False)
        im.voxelsize = readheader(filename).get(
            "pixelspacing_tczyx",
            [
                1.0,
            ]
            * 5,
        )[:1:-1]
    else:
        im = tmp.transpose(2, 1, 0)
    return im


def write_klb(filename, im):
    if np.isfortran(im):
        if len(im.shape) == 3:
            writefull(
                im.transpose(2, 1, 0),
                filename,
                pixelspacing_tczyx=(1, 1) + tuple(im.resolution[::-1]),
            )
        else:
            writefull(
                im.transpose(1, 0),
                filename,
                pixelspacing_tczyx=(1, 1) + tuple(im.resolution[::-1]),
            )
    else:
        writefull(im, filename)



import tifffile
import os
import registrationtools
import json
import numpy as np
from glob import glob
from ensure import check ##need to pip install !
from pathlib import Path

def get_paths() :
#Ask the user for the folder containing his data, find the movies is they are in tiff format and put them in a list after confirmation
    path_correct=False
    while path_correct==False :
        path_folder = input('Path to the folder of the movie(s) (in tiff format only) : \n ')
        paths_movies = sorted(glob(rf'{path_folder}/*.tif')) 
        if len(paths_movies) > 0 :
            print('You have',len(paths_movies),'movie(s), which is (are) : ')
            for path_movie in paths_movies :
                print('',Path(path_movie).stem) #the empty str at the begining is to align the answers one space after the questions, for readability
                path_correct = int(input('Correct ? (1 for yes, 0 for no) \n '))
        else :
            print('There is no tiff file in the folder. \n')

    return (paths_movies)

def dimensions(list_paths:list) :
    # Take the path to the data and returns its size in the dimensions CTZXY. Raise error if the other movies do not have the same C ot T dimensions.
    movie = tifffile.imread(list_paths[0])
    name_movie = Path(list_paths[0]).stem
    dim_correct = False
    while dim_correct == False :
        print('\nThe dimensions of ',name_movie, 'are ',movie.shape,'. \n')
        dimensions=input('What is the order of the dimensions (for example TZCYX or XYZT) ? T stands for Time, C for channels if your image has multiple channels, Z for depth (or number or plans) and XY is your field of view. \n ')
        sorted_axes = sorted(dimensions)
        if ''.join(sorted_axes) == "CTXYZ":
            number_channels = movie.shape[dimensions.find('C')]
            dim_correct=True
        elif ''.join(sorted_axes) == "TXYZ":
            number_channels = 1
            dim_correct=True
        elif len(sorted_axes) != len(movie.shape) :
            print('Error : Number of dimensions is incorrect. \n')
            dim_correct = False 
        else:
            print(' \n The letters you choose has to be among these letters : X,Y,Z,C,T, with XYZT mandatory, and no other letters are allowed. \nEvery letter can be included only once. \n')

        number_timepoints = movie.shape[dimensions.find('T')]

        if 'C' in dimensions :
            number_channels = movie.shape[dimensions.find('C')]
        else :
            number_channels = 1
            
        if dim_correct == True :
            depth = movie.shape[dimensions.find('Z')]
            size_X = movie.shape[dimensions.find('X')]
            size_Y = movie.shape[dimensions.find('Y')]

            print('\nSo',name_movie, 'has',number_channels,'channels, ',number_timepoints,'timepoints, the depth in z is',depth,'pixels and the XY plane measures ',size_X,'x',size_Y,'pixels.')
            dim_correct = int(input('Correct ? (1 for yes, 0 for no) \n '))

    #checking if the movies have the same dimensions in C and T, otherwise the registration cannot be computed.
    if 'C' in dimensions :    #have to check beforehand if multichannels, otherwise, dimensions.find('C') will return -1 and mov.shape[dimensions.find('C')] will probably return the dimension of the X axis
        for path in list_paths : 
            mov = tifffile.imread(path)
            check(mov.shape[dimensions.find('T')]).equals(number_timepoints).or_raise(Exception, 'These movies do not all have the same number of timepoints. \n')
            check(mov.shape[dimensions.find('C')]).equals(number_channels).or_raise(Exception, 'These movies do not all have the same number of channels. \n')
            #if XYZ are not the same thats ok ?
    return(number_channels,number_timepoints,depth,size_X,size_Y)


def get_channels_name(number_channels:int) :
#Ask for the name of the channels, return a list of str containing all the channels
    channels=[]
    if number_channels == 1:
        ch = str(input('Name of the channel : \n '))
        channels.append(ch)
    else : #if multichannels
        for n in range(number_channels) :
            ch = str(input('Name of channel n°'+str(n+1) +' : \n '))
            channels.append(ch)
    return(channels)

def reference_channel(channels:list) :
#Ask for the reference channel among the channels given (with safety check)
    if len(channels) == 1 :
        ch_ref=channels[0]
    else :
        channel_correct=False
        print('\nAmong the channels'+ str(channels)+', you need a reference channel to compute the registration. A good option is generally a marker that is expressed ubiquitously\n')
        while channel_correct==False :
            ch_ref = input('Name of the reference channel : \n ')
            if ch_ref not in channels :
                print('The reference channel is not in', channels,'(do not put any other character than the name itself)')
                channel_correct=False
            else :
                channel_correct=True
    channels_float=channels.copy()
    channels_float.remove(ch_ref)
    return(channels_float,ch_ref)



def sort_by_channels_and_timepoints(list_paths:str,channels:str,number_timepoints:int):
#Take a list of movies and cut them into a timesequence of 3D stacks, one timesequence per channel. Works for one channel or multiple channels.
    for path in list_paths :
        for n in range(len(channels)) :
            name_movie = Path(path).stem
            directory = name_movie+'_'+channels[n]
            cut_timesequence(path_to_movie = path,
                                directory=directory,
                                ind_current_channel=n,
                                number_timepoints = number_timepoints,
                                number_channels = len(channels))


def cut_timesequence (path_to_movie:str, directory:str, ind_current_channel:int, number_timepoints:int,number_channels:int):
    #take a movie and cut it into a timesequence in the given directory. Creates the folder structure for the registration.
    movie = tifffile.imread(path_to_movie)
    position_t = np.argwhere(np.array(movie.shape)==number_timepoints)[0][0]
    movie=np.moveaxis(movie,source=position_t,destination=0) #i did not test this
    path_to_data=os.path.dirname(path_to_movie)
    path_dir = rf'{path_to_data}\{directory}'
    check(os.path.isdir(path_dir)).is_(False).or_raise(Exception,"Please delete the folder "+directory+" and run again.")
    os.mkdir(os.path.join(path_to_data,directory))
    os.mkdir(os.path.join(path_to_data+'/'+directory,"trsf"))
    os.mkdir(os.path.join(path_to_data+'/'+directory,"output"))
    os.mkdir(os.path.join(path_to_data+'/'+directory,"proj_output"))
    os.mkdir(os.path.join(path_to_data+'/'+directory,"stackseq"))
    
    if number_channels >1 :
        position_c = np.argwhere(np.array(movie.shape)==number_channels)[0][0]
        movie=np.moveaxis(movie,source=position_c,destination=2) #we artificially change to the format TZCXY (reference in Fiji). Its just to cut into timesequence. does not modify the data
        for t in range(number_timepoints) :
            stack = movie[t,:,ind_current_channel,:,:]
            tifffile.imwrite(path_to_data+'/'+directory+"/stackseq/movie_t"+str(format(t,'03d')+'.tif'),stack)
    else : 
        for t in range(number_timepoints) :
            stack = movie[t,:,:,:]
            tifffile.imwrite(path_to_data+'/'+directory+"/stackseq/movie_t"+str(format(t,'03d')+'.tif'),stack)

def get_voxel_sizes() :
#ask the user for the voxel size of his input and what voxel size does he want in output. Returns 2 tuples for this values.
    print('\nTo register properly, you need to specify the voxel size of your input image. This can be found in Fiji, Image>Show Info. ')
    print('Voxel size of your original image (XYZ successively) :\n ')
    x= float(input('X :'))
    y= float(input('Y :'))
    z= float(input('Z :'))
    voxel_size_input = [x,y,z]
    print('Initial voxel size =',voxel_size_input)

    change_voxel_size = int(input(' \nYou can choose to have another voxel size on the registered image , for example to have an isotropic output image (voxel size [1,1,1]), Or you can also choose to keep the same voxel size. \nDo you want to change the voxel size of your movies ? (1 for yes, 0 for no) : \n '))
    if change_voxel_size==1 :
        print('\nVoxel size of your image after transformation (XYZ): \n ')
        x= float(input('X :'))
        y= float(input('Y :'))
        z= float(input('Z :'))
        voxel_size_output = [x,y,z]
    elif change_voxel_size==0 :
        voxel_size_output = voxel_size_input
    print('\nVoxel size after transformation =',voxel_size_output)

    return(voxel_size_input,voxel_size_output)


def get_trsf_type() :
#Ask for what transformation the user want and return a string
    list_trsf_types = ['rigid2D','rigid3D','translation2D','translation3D'] #needs to be completed
    print('\nYou can choose to apply different transformation types depending on your data : ',list_trsf_types)
    trsf_correct = False
    while trsf_correct == False :
        trsf_type = str(input('\nWhich one do you want to use ? (please enter the name of the transformation only, no other character) \n '))
        if trsf_type in list_trsf_types :
            trsf_correct = True  
        else :
            print(('You can only choose a transformation that is in this list :',list_trsf_types))

    return(trsf_type)


def data_preparation() :
#Gather all the functions asking the user for the parameters. Returns the useful ones for registration (not XYZ size and channels list)
    list_paths = get_paths()
    number_channels,number_timepoints,depth,size_X,size_Y = dimensions(list_paths)  
    channels=get_channels_name(number_channels)
    channels_float,ch_ref = reference_channel(channels)
    
    sort_by_channels_and_timepoints(list_paths = list_paths,
                                    channels = channels,
                                    number_timepoints = number_timepoints)
    
    voxel_size_input, voxel_size_output = get_voxel_sizes()
    trsf_type = get_trsf_type()

    return(list_paths,
           number_timepoints,
           channels_float,ch_ref,
           voxel_size_input, voxel_size_output,
           trsf_type)


def run_registration (list_paths:list,channels_float:list, ch_ref:str,  voxel_size_input:tuple, voxel_size_output:tuple, trsf_type:str) :
#Does the actual registration : Take the reference channel first to compute and apply the trsf, then loop on the float channels and only apply it
    for path in list_paths :
        json_string=[]
        folder=os.path.dirname(path)
        name_movie = Path(path).stem
        movie = tifffile.imread(path)
        directory = name_movie+'_'+ch_ref
        data_ref = {
                "path_to_data": rf'{folder}/{directory}/stackseq/',
                "file_name": "movie_t{t:03d}.tif",
                "trsf_folder": rf'{folder}/{directory}/trsf/',
                "output_format": rf'{folder}/{directory}/output/',
                "projection_path":rf'{folder}/{directory}/proj_output/',
                "check_TP": 0,
                "voxel_size": voxel_size_input,
                "voxel_size_out" : voxel_size_output,
                "first": 0,
                "last": movie.shape[0]-1,
                "not_to_do": [],
                "compute_trsf": 1,
                "ref_TP": int(movie.shape[0]/2),
                "trsf_type": trsf_type,
                "padding": 1,
                "recompute": 1,
                "apply_trsf":1,
                "out_bdv": "",
                "plot_trsf":0
                }
        json_string.append(json.dumps(data_ref))
        tr = registrationtools.TimeRegistration(data_ref)
        tr.run_trsf() 

        #registration rest of the channels
        for c in channels_float :
            directory = name_movie+'_'+c
            data_float = {
                    "path_to_data": rf'{folder}/{directory}/stackseq/',
                    "file_name": "movie_t{t:03d}.tif",
                    "trsf_folder": rf'{folder}/{name_movie}_{ch_ref}/trsf/',
                    "output_format": rf'{folder}/{directory}/output/',
                    "projection_path":rf'{folder}/{directory}/proj_output/',
                    "check_TP": 0,
                    "voxel_size": voxel_size_input,
                    "voxel_size_out" : voxel_size_output,
                    "first": 0,
                    "last": movie.shape[0]-1,
                    "not_to_do": [],
                    "compute_trsf": 0,
                    "ref_TP": int(movie.shape[0]/2),
                    "trsf_type": trsf_type,
                    "padding": 1,
                    "recompute": 1,
                    "apply_trsf":1,
                    "out_bdv": "",
                    "plot_trsf":0
                    }
            json_string.append(json.dumps(data_float))
            tr = registrationtools.TimeRegistration(data_float)
            tr.run_trsf()
    return(json_string)



def save_sequences_as_stacks(list_paths:list,channels:list,number_timepoints:int) :
    #save the timesequence as a hyperstack
    for path in list_paths :
        path_to_data=os.path.dirname(path)
        name_movie = Path(path).stem
        movie = tifffile.imread(path)
        stack0 =tifffile.imread(rf"{path_to_data}/{name_movie}_{channels[0]}/output/movie_t000.tif") #we use the first image (3D) to know the dimensions
        registered_movie = np.zeros((number_timepoints,stack0.shape[0],len(channels),stack0.shape[1],stack0.shape[2]),dtype=np.float32) #one movie per channel, of format (t,z,y,x).Datatype uint16 or float32 is necessary to export as hyperstack
        for ind_c,c in enumerate(channels) :
            directory = name_movie+'_'+c
            for t in range(movie.shape[0]) :
                stack =tifffile.imread(rf"{path_to_data}/{directory}/output/movie_t{format(t,'03d')}.tif")
                #we take each stack in a given timepoint
                registered_movie[t,:,ind_c,:,:]=stack #and put it in a new hyperstack
        tifffile.imwrite(path_to_data+rf"/{name_movie}_registered.tif",registered_movie.astype(np.float32),imagej=True) #write a hyperstack in the main folder
        print('saved registered', name_movie, 'of size', registered_movie.shape)


def save_jsonfile(list_paths,json_string):
    path_to_data = os.dirname(list_paths[0])
    keep_same_dir = int(input(str('Do you want to save your json files in the same master directory, in '+path_to_data+'\jsonfiles ? (1 for yes, 0 for no)')))
    if keep_same_dir ==1 :
        path_to_json = rf'{path_to_data}\jsonfiles'
        os.mkdir(os.path.join(path_to_data,'jsonfiles'))
    else :
        path_to_json = input('In which folder do you want to write your jsonfile ?').replace("\\", "/") #if single backslashes, fine. If double backslashes (when copy/paste the Windows path), compatibility problems, thats why we replace by single slashes.')
    
    print('saving',len(json_string),'json files :') 
    for ind_json,json in enumerate(json_string) :
        with open(path_to_json+'\param'+str(ind_json)+'.json','w') as outfile : #maybe not the best name
            outfile.write(json)
            print(path_to_json)
    print('Done saving')
    


from h5py import File
from .spatial_image import SpatialImage
import numpy as np


def read_h5(filename):
    """Read an hdf5 file

    :Parameters:
     - `filename` (str) - name of the file to read
    """
    data = File(filename.replace("\\", ""), "r")["Data"]
    im_out = np.zeros(data.shape, dtype=data.dtype)
    data.read_direct(im_out)
    return SpatialImage(im_out.transpose(2, 1, 0))


def write_h5(filename, im):
    """Write an image to an hdf5

    :Parameters:
     - `filename` (str) - name of the file to write
     - `im` SpatialImage - image to write
    """
    f = File(filename, mode="w")
    dset = f.create_dataset("Data", data=im.transpose(2, 1, 0), chunks=True)
    f.close()



__version__ = "0.4.3"

from .time_registration import TimeRegistration, time_registration
from .spatial_registration import SpatialRegistration, spatial_registration
from .utils import *
import sys

if 8 < sys.version_info.minor:
    import importlib.resources as importlib_resources

    pkg = importlib_resources.files("registrationtools") / "data"
else:
    from importlib_resources._legacy import path

    pkg = path("registrationtools", "data").args[0]

image_path = pkg / "images"
json_path_spatial = pkg / "JSON_spatial"
json_path_time = pkg / "JSON_time"



from os.path import (
    exists,
    splitext,
    isdir,
    split as psplit,
    expanduser as expusr,
)
import os, fnmatch
import warnings
from struct import pack, unpack, calcsize
from pickle import dumps, loads
import numpy as np

from .spatial_image import SpatialImage

from .inrimage import read_inrimage, write_inrimage

try:
    from .tif import read_tif, write_tif
except Exception as e:
    warnings.warn("pylibtiff library is not installed")

try:
    from .h5 import read_h5, write_h5
except Exception as e:
    warnings.warn("h5py library is not installed")

try:
    from .klb import read_klb, write_klb
except Exception as e:
    warnings.warn("KLB library is not installed")

from .folder import read_folder


def imread(filename, parallel=True, SP_im=True):
    """Reads an image file completely into memory.

    It uses the file extension to determine how to read the file. It first tries
    some specific readers for volume images (Inrimages, TIFFs, LSMs, NPY) or falls
    back on PIL readers for common formats if installed.

    In all cases the returned image is 3D (2D is upgraded to single slice 3D).
    If it has colour or is a vector field it is even 4D.

    :Parameters:
     - `filename` (str)

    :Returns Type:
        |SpatialImage|
    """
    filename = expusr(filename)
    if not exists(filename):
        raise IOError("The requested file do not exist: %s" % filename)

    root, ext = splitext(filename)
    ext = ext.lower()
    if ext == ".gz":
        root, ext = splitext(root)
        ext = ext.lower()
    if ext == ".inr":
        return read_inrimage(filename)
    elif ext in [".tif", ".tiff"]:
        return read_tif(filename)
    elif ext in [".h5", ".hdf5"]:
        return read_h5(filename)
    elif ext == ".klb":
        return read_klb(filename, SP_im)
    elif isdir(filename):
        return read_folder(filename, parallel)


def imsave(filename, img):
    """Save a |SpatialImage| to filename.

    .. note: `img` **must** be a |SpatialImage|.

    The filewriter is choosen according to the file extension. However all file extensions
    will not match the data held by img, in dimensionnality or encoding, and might raise `IOError`s.

    For real volume data, Inrimage and NPY are currently supported.
    For |SpatialImage|s that are actually 2D, PNG, BMP, JPG among others are supported if PIL is installed.

    :Parameters:
     - `filename` (str)
     - `img` (|SpatialImage|)
    """

    filename = expusr(filename)
    root, ext = splitext(filename)

    # assert isinstance(img, SpatialImage) or ext == '.klb'
    # -- images are always at least 3D! If the size of dimension 3 (indexed 2) is 1, then it is actually
    # a 2D image. If it is 4D it has vectorial or RGB[A] data. --
    head, tail = psplit(filename)
    head = head or "."
    if not exists(head):
        raise IOError("The directory do not exist: %s" % head)

    # is2D = img.shape[2] == 1
    ext = ext.lower()
    if ext == ".gz":
        root, ext = splitext(root)
        ext = ext.lower()
    if ext == ".inr":
        write_inrimage(filename, img)
    elif ext in [".tiff", ".tif"]:
        write_tif(filename, img)
    elif ext == ".klb":
        write_klb(filename, img)
    elif ext in [".h5", ".hdf5"]:
        write_h5(filename, img)



# -*- python -*-
#
#       image.serial: read/write spatial nd images
#
#       Copyright 2006 - 2011 INRIA - CIRAD - INRA
#
#       File author(s): Jerome Chopard <jerome.chopard@sophia.inria.fr>
#                       Eric Moscardi <eric.moscardi@sophia.inria.fr>
#                       Daniel Barbeau <daniel.barbeau@sophia.inria.fr>
#
#       Distributed under the Cecill-C License.
#       See accompanying file LICENSE.txt or copy at
#           http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html
#
#       OpenAlea WebSite : http://openalea.gforge.inria.fr
################################################################################
"""
This module defines inrimage format
"""

__license__ = "Cecill-C"
__revision__ = " $Id$ "

import os
from os import path
import numpy as np
from struct import calcsize, pack, unpack
import gzip
from io import BytesIO
from .spatial_image import SpatialImage

__all__ = ["read_inriheader", "read_inrimage", "write_inrimage"]

specific_header_keys = (
    "XDIM",
    "YDIM",
    "ZDIM",
    "VDIM",
    "TYPE",
    "PIXSIZE",
    "SCALE",
    "CPU",
    "VX",
    "VY",
    "VZ",
    "TX",
    "TY",
    "TZ",
    "#GEOMETRY",
)


def open_inrifile(filename):
    """Open an inrimage file

    Manage the gz attribute
    """
    if path.splitext(filename)[1] in (".gz", ".zip"):
        with gzip.open(filename, "rb") as fzip:
            f = BytesIO(fzip.read())
            fzip.close()
    else:
        f = open(filename, "rb")
    return f


def _read_header(f):
    """Extract header from a stream and return it
    as a python dict
    """
    # read header string
    header = ""
    while header[-4:] != "##}\n":
        header += f.read(256).decode("latin-1")

    # read infos in header
    prop = {}
    hstart = header.find("{\n") + 1
    hend = header.find("##}")
    infos = [gr for gr in header[hstart:hend].split("\n") if len(gr) > 0]

    # format infos
    for prop_def in infos:
        key, val = prop_def.split("=")
        prop[key] = val

    # return
    return prop


def read_inriheader(filename):
    """Read only the header of an inrimage

    :Parameters:
     - `filename` (str) - name of the file to read
    """
    f = open_inrifile(filename)
    prop = _read_header(f)
    f.close()
    return prop


def read_inrimage(filename):
    """Read an inrimage, either zipped or not according to extension

    :Parameters:
     - `filename` (str) - name of the file to read
    """
    f = open_inrifile(filename)

    # read header
    prop = _read_header(f)
    prop["filename"] = filename  # Jonathan : 14.05.2012

    # extract usefull infos to read image
    zdim = int(prop.pop("ZDIM"))
    ydim = int(prop.pop("YDIM"))
    xdim = int(prop.pop("XDIM"))
    vdim = int(prop.pop("VDIM", 1))
    # if vdim != 1 :
    #   msg = "don't know how to handle vectorial pixel values"
    #   raise NotImplementedError(msg)

    # find data type
    pixsize = int(prop.pop("PIXSIZE", "0").split(" ")[0])
    dtype = prop.pop("TYPE")

    if dtype == "unsigned fixed":
        if pixsize == 0:
            ntyp = np.dtype(np.int)
        else:
            try:
                ntyp = eval("np.dtype(np.uint%d)" % pixsize)
            except AttributeError:
                raise UserWarning("undefined pix size: %d" % pixsize)
    elif dtype == "float":
        if pixsize == 0:
            ntyp = np.dtype(np.float)
        else:
            try:
                ntyp = eval("np.dtype(np.float%d)" % pixsize)
            except AttributeError:
                raise UserWarning("undefined pix size: %d" % pixsize)
    else:
        msg = "unable to read that type of datas : %s" % dtype
        raise UserWarning(msg)

    # read datas
    size = ntyp.itemsize * xdim * ydim * zdim * vdim
    mat = np.frombuffer(f.read(size), ntyp)
    if vdim != 1:
        mat = mat.reshape((vdim, xdim, ydim, zdim), order="F")
        mat = mat.transpose(1, 2, 3, 0)
    else:
        mat = mat.reshape((xdim, ydim, zdim), order="F")
        # mat = mat.transpose(2,1,0)

    # create SpatialImage
    res = tuple(float(prop.pop(k)) for k in ("VX", "VY", "VZ"))
    for k in ("TX", "TY", "TZ"):
        prop.pop(k, None)

    img = SpatialImage(mat, res, vdim, prop)

    # return
    f.close()
    return img


def write_inrimage_to_stream(stream, img):
    assert img.ndim in (3, 4)

    # metadata
    info = dict(getattr(img, "info", {}))
    if "filename" in info:
        info["filename"] = path.basename(info["filename"])
    else:
        info["filename"] = path.basename("img.inr")

    # image dimensions
    if img.ndim < 4:
        info["XDIM"], info["YDIM"], info["ZDIM"] = (
            "%d" % val for val in img.shape
        )
        info["VDIM"] = "1"
    else:
        info["XDIM"], info["YDIM"], info["ZDIM"], info["VDIM"] = (
            "%d" % val for val in img.shape
        )

    # image resolution
    res = getattr(img, "resolution", (1, 1, 1))
    info["VX"], info["VY"], info["VZ"] = ("%f" % val for val in res)

    # data type
    if img.dtype == np.uint8:
        info["TYPE"] = "unsigned fixed"
        info["PIXSIZE"] = "8 bits"
    elif img.dtype == np.uint16:
        info["TYPE"] = "unsigned fixed"
        info["PIXSIZE"] = "16 bits"
    elif img.dtype == np.uint32:
        info["TYPE"] = "unsigned fixed"
        info["PIXSIZE"] = "32 bits"
    elif img.dtype == np.uint64:
        info["TYPE"] = "unsigned fixed"
        info["PIXSIZE"] = "64 bits"
    elif img.dtype == np.float32:
        info["TYPE"] = "float"
        info["PIXSIZE"] = "32 bits"
    elif img.dtype == np.float64:
        info["TYPE"] = "float"
        info["PIXSIZE"] = "64 bits"
    # elif img.dtype == np.float128 :
    #   info["TYPE"] = "float"
    #   info["PIXSIZE"] = "128 bits"
    else:
        msg = "unable to write that type of datas : %s" % str(img.dtype)
        raise UserWarning(msg)

    # mandatory else an error occurs when reading image
    info["#GEOMETRY"] = "CARTESIAN"
    info["CPU"] = "decm"

    # write header
    header = "#INRIMAGE-4#{\n"
    for (
        k
    ) in (
        specific_header_keys
    ):  # HACK pas bo to ensure order of specific headers
        try:
            header += "%s=%s\n" % (k, info[k])
        except KeyError:
            pass

    for k in set(info) - set(specific_header_keys):
        header += "%s=%s\n" % (k, info[k])

    # fill header to be a multiple of 256
    header_size = len(header) + 4
    if (header_size % 256) > 0:
        header += "\n" * (256 - header_size % 256)

    header += "##}\n"

    contiguous_rule = "C" if img.flags["C_CONTIGUOUS"] else "F"
    stream.write(header)
    if img.ndim == 3:
        stream.write(img.tostring(contiguous_rule))
    elif img.ndim == 4:
        mat = img.transpose(3, 0, 1, 2)
        stream.write(mat.tostring(contiguous_rule))
    else:
        raise Exception("Unhandled image dimension %d." % img.ndim)


def write_inrimage(filename, img):
    """Write an inrimage zipped or not according to the extension

    .. warning:: if img is not a |SpatialImage|, default values will be used
                 for the resolution of the image

    :Parameters:
     - `img` (|SpatialImage|) - image to write
     - `filename` (str) - name of the file to read
    """
    # open stream
    zipped = path.splitext(filename)[1] in (".gz", ".zip")

    if zipped:
        f = gzip.GzipFile(filename, "wb")
        # f = StringIO()
    else:
        f = open(filename, "wb")

    try:
        write_inrimage_to_stream(f, img)
    except:
        # -- remove probably corrupt file--
        f.close()
        if path.exists(filename) and path.isfile(filename):
            os.remove(filename)
        raise
    else:
        f.close()



# -*- python -*-
#
#       adapted from C. Pradal
#
#       Copyright 2011 INRIA - CIRAD - INRA
#
#       File author(s): Christophe Pradal <christophe.pradal@cirad.fr>
#                       Daniel Barbeau    <daniel.barbeau@cirad.fr>
#
#       Distributed under the Cecill-C License.
#       See accompanying file LICENSE.txt or copy at
#           http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html
#
#       OpenAlea WebSite : http://openalea.gforge.inria.fr
################################################################################
"""
This module reads 3D tiff format
"""


__license__ = "Cecill-C"
__revision__ = " $Id$ "

import numpy as np
from .spatial_image import SpatialImage


__all__ = []

import decimal

import numpy as np
from tifffile import TiffFile, imread, imwrite
import os, os.path, sys, time

__all__ += ["read_tif", "write_tif", "mantissa"]


def read_tif(filename, channel=0):
    """Read a tif image

    :Parameters:
    - `filename` (str) - name of the file to read
    """

    def format_digit(v):
        try:
            v = eval(v)
        except Exception as e:
            pass
        return v

    with TiffFile(filename) as tif:
        if "ImageDescription" in tif.pages[0].tags:
            description = (
                tif.pages[0].tags["ImageDescription"].value.split("\n")
            )
            separator = set.intersection(*[set(k) for k in description]).pop()
            info_dict = {
                v.split(separator)[0]: format_digit(v.split(separator)[1])
                for v in description
            }
        else:
            info_dict = {}

        if "XResolution" in tif.pages[0].tags:
            vx = tif.pages[0].tags["XResolution"].value
            if vx[0] != 0:
                vx = vx[1] / vx[0]
                if isinstance(vx, list):
                    vx = vx[1] / vx[0]
            else:
                vx = 1.0
        else:
            vx = 1.0

        if "YResolution" in tif.pages[0].tags:
            vy = tif.pages[0].tags["YResolution"].value
            if vy[0] != 0:
                vy = vy[1] / vy[0]
                if isinstance(vy, list):
                    vy = vy[1] / vy[0]
            else:
                vy = 1.0
        else:
            vy = 1.0

        if "ZResolution" in tif.pages[0].tags:
            vz = tif.pages[0].tags["ZResolution"].value
            if vz[0] != 0:
                if isinstance(vz, list):
                    vz = vz[1] / vz[0]
            else:
                vz = 1.0
        elif "spacing" in info_dict:
            vz = info_dict["spacing"]
        else:
            vz = 1.0

        im = tif.asarray()
        if len(im.shape) == 3:
            im = np.transpose(im, (2, 1, 0))
        elif len(im.shape) == 2:
            im = np.transpose(im, (1, 0))
        if 3 <= len(im.shape):
            im = SpatialImage(im, (vx, vy, vz))
        else:
            print(im.shape, vx, vy)
            im = SpatialImage(im, (vx, vy))
        im.resolution = (vx, vy, vz)
    return im


def mantissa(value):
    """Convert value to [number, divisor] where divisor is power of 10"""
    # -- surely not the nicest thing around --
    d = decimal.Decimal(str(value))  # -- lovely...
    sign, digits, exp = d.as_tuple()
    n_digits = len(digits)
    dividend = int(
        sum(v * (10 ** (n_digits - 1 - i)) for i, v in enumerate(digits))
        * (1 if sign == 0 else -1)
    )
    divisor = int(10**-exp)
    return dividend, divisor


def write_tif(filename, obj):
    if len(obj.shape) > 3:
        raise IOError(
            "Vectorial images are currently unsupported by tif writer"
        )
    is3D = len(obj.shape) == 3

    if hasattr(obj, "resolution"):
        res = obj.resolution
    elif hasattr(obj, "voxelsize"):
        res = obj.resolution
    else:
        res = (1, 1, 1) if is3D else (1, 1)

    vx = 1.0 / res[0]
    vy = 1.0 / res[1]
    if is3D:
        spacing = res[2]
        metadata = {"spacing": spacing, "axes": "ZYX"}
    else:
        metadata = {}
    extra_info = {
        "XResolution": res[0],
        "YResolution": res[1],
        "spacing": spacing
        if is3D
        else None,  # : no way to save the spacing (no specific tag)
    }
    print(extra_info)

    if obj.dtype.char in "BHhf":
        return imwrite(
            filename,
            obj.T,
            imagej=True,
            resolution=(vx, vy),
            metadata=metadata,
        )
    else:
        return imwrite(filename, obj.T)



#!/usr/bin/python
# This file is subject to the terms and conditions defined in
# file 'LICENCE', which is part of this source code package.
# Author: Leo Guignard (leo.guignard...@AT@...univ-amu.fr)

from pathlib import Path
from time import time
import os
from subprocess import call
import scipy as sp
from scipy import interpolate
import json
import numpy as np
from IO import imread, imsave, SpatialImage
from typing import List, Tuple
from statsmodels.nonparametric.smoothers_lowess import lowess
import sys
import xml.etree.ElementTree as ET
from xml.dom import minidom
from transforms3d.affines import decompose
from transforms3d.euler import mat2euler

if sys.version_info[0] < 3:
    from future.builtins import input

try:
    from pyklb import readheader

    pyklb_found = True
except Exception as e:
    pyklb_found = False
    print("pyklb library not found, klb files will not be generated")


class trsf_parameters(object):
    """
    Read parameters for the registration function from a preformated json file
    """

    def check_parameters_consistancy(self) -> bool:
        """
        Function that should check parameter consistancy
        """

        correct = True
        if not "path_to_data" in self.__dict__:
            print('\n\t"path_to_data" is required')
            correct = False
        if not "file_name" in self.__dict__:
            print('\n\t"file_name" is required')
            correct = False
        if not "trsf_folder" in self.__dict__:
            print('\n\t"trsf_folder" is required')
            correct = False
        if not "voxel_size" in self.__dict__:
            print('\n\t"voxel_size" is required')
            correct = False
        if not "ref_TP" in self.__dict__:
            print('\n\t"ref_TP" is required')
            correct = False
        if not "projection_path" in self.__dict__:
            print('\n\t"projection_path" is required')
            correct = False
        if self.apply_trsf and not "output_format" in self.__dict__:
            print('\n\t"output_format" is required')
            correct = False
        if self.trsf_type != "translation" and (
            self.lowess or self.trsf_interpolation
        ):
            print("\n\tLowess or transformation interplolation")
            print("\tonly work with translation")
            correct = False
        if self.lowess and not "window_size" in self.param_dict:
            print('\n\tLowess smoothing "window_size" is missing')
            print("\tdefault value of 5 will be used\n")
        if self.trsf_interpolation and not "step_size" in self.param_dict:
            print('\n\tTransformation interpolation "step_size" is missing')
            print("\tdefault value of 100 will be used\n")
        if self.trsf_type == "vectorfield" and self.ref_path is None:
            print("\tNon-linear transformation asked with propagation.")
            print("\tWhile working it is highly not recommended")
            print(
                "\tPlease consider not doing a registration from propagation"
            )
            print("\tTHIS WILL LITERALLY TAKE AGES!! IF IT WORKS AT ALL ...")
        if (
            not isinstance(self.spline, int)
            or self.spline < 1
            or 5 < self.spline
        ):
            out = ("{:d}" if isinstance(self.spline, int) else "{:s}").format(
                self.spline
            )
            print("The degree of smoothing for the spline interpolation")
            print("Should be an Integer between 1 and 5, you gave " + out)
        return correct

    def __str__(self) -> str:
        max_key = (
            max([len(k) for k in self.__dict__.keys() if k != "param_dict"])
            + 1
        )
        max_tot = (
            max(
                [
                    len(str(v))
                    for k, v in self.__dict__.items()
                    if k != "param_dict"
                ]
            )
            + 2
            + max_key
        )
        output = "The registration will run with the following arguments:\n"
        output += "\n" + " File format ".center(max_tot, "-") + "\n"
        output += "path_to_data".ljust(max_key, " ") + ": {}\n".format(
            self.path_to_data
        )
        output += "file_name".ljust(max_key, " ") + ": {}\n".format(
            self.file_name
        )
        output += "trsf_folder".ljust(max_key, " ") + ": {}\n".format(
            self.trsf_folder
        )
        output += "output_format".ljust(max_key, " ") + ": {}\n".format(
            self.output_format
        )
        output += "check_TP".ljust(max_key, " ") + ": {:d}\n".format(
            self.check_TP
        )
        output += "\n" + " Time series properties ".center(max_tot, "-") + "\n"
        output += "voxel_size".ljust(
            max_key, " "
        ) + ": {:f}x{:f}x{:f}\n".format(*self.voxel_size)
        output += "first".ljust(max_key, " ") + ": {:d}\n".format(self.first)
        output += "last".ljust(max_key, " ") + ": {:d}\n".format(self.last)
        output += "\n" + " Registration ".center(max_tot, "-") + "\n"
        output += "compute_trsf".ljust(max_key, " ") + ": {:d}\n".format(
            self.compute_trsf
        )
        if self.compute_trsf:
            output += "ref_TP".ljust(max_key, " ") + ": {:d}\n".format(
                self.ref_TP
            )
            if self.ref_path is not None:
                output += "ref_path".ljust(max_key, " ") + ": {}\n".format(
                    self.ref_path
                )
            output += "trsf_type".ljust(max_key, " ") + ": {}\n".format(
                self.trsf_type
            )
            if self.trsf_type == "vectorfield":
                output += "sigma".ljust(max_key, " ") + ": {:.2f}\n".format(
                    self.sigma
                )
            output += "padding".ljust(max_key, " ") + ": {:d}\n".format(
                self.padding
            )
            output += "lowess".ljust(max_key, " ") + ": {:d}\n".format(
                self.lowess
            )
            if self.lowess:
                output += "window_size".ljust(
                    max_key, " "
                ) + ": {:d}\n".format(self.window_size)
            output += "trsf_interpolation".ljust(
                max_key, " "
            ) + ": {:d}\n".format(self.trsf_interpolation)
            if self.trsf_interpolation:
                output += "step_size".ljust(max_key, " ") + ": {:d}\n".format(
                    self.step_size
                )
            output += "recompute".ljust(max_key, " ") + ": {:d}\n".format(
                self.recompute
            )
        output += "apply_trsf".ljust(max_key, " ") + ": {:d}\n".format(
            self.apply_trsf
        )
        if self.apply_trsf:
            if self.projection_path is not None:
                output += "projection_path".ljust(
                    max_key, " "
                ) + ": {}\n".format(self.projection_path)
            output += "image_interpolation".ljust(
                max_key, " "
            ) + ": {}\n".format(self.image_interpolation)

        return output

    def add_path_prefix(self, prefix: str):
        """
        Add a prefix to all the relevant paths (this is mainly for the unitary tests)

        Args:
            prefix (str): The prefix to add in front of the path
        """
        self.projection_path = (
            os.path.join(prefix, self.projection_path) + os.path.sep
        )
        self.path_to_data = (
            os.path.join(prefix, self.path_to_data) + os.path.sep
        )
        self.trsf_folder = os.path.join(prefix, self.trsf_folder) + os.path.sep
        self.output_format = (
            os.path.join(prefix, self.output_format) + os.path.sep
        )

    def __init__(self, file_name: str):
        if not isinstance(file_name, dict):
            with open(file_name) as f:
                param_dict = json.load(f)
                f.close()
        else:
            param_dict = {}
            for k, v in file_name.items():
                if isinstance(v, Path):
                    param_dict[k] = str(v)
                else:
                    param_dict[k] = v

        # Default parameters
        self.check_TP = None
        self.not_to_do = []
        self.compute_trsf = True
        self.ref_path = None
        self.padding = 1
        self.lowess = False
        self.window_size = 5
        self.step_size = 100
        self.recompute = True
        self.apply_trsf = True
        self.sigma = 2.0
        self.keep_vectorfield = False
        self.trsf_type = "rigid"
        self.image_interpolation = "linear"
        self.path_to_bin = ""
        self.spline = 1
        self.trsf_interpolation = False
        self.sequential = True
        self.time_tag = "TM"
        self.do_bdv = 0
        self.bdv_voxel_size = None
        self.bdv_unit = "microns"
        self.projection_path = None
        self.pre_2D = False
        self.low_th = None
        self.plot_trsf = False

        self.param_dict = param_dict
        if "registration_depth" in param_dict:
            self.__dict__["registration_depth_start"] = 6
            self.__dict__["registration_depth_end"] = param_dict[
                "registration_depth"
            ]
        elif not "registration_depth_start" in param_dict:
            self.__dict__["registration_depth_start"] = 6
            self.__dict__["registration_depth_end"] = 3

        self.__dict__.update(param_dict)
        self.voxel_size = tuple(self.voxel_size)
        if not hasattr(self, "voxel_size_out"):
            self.voxel_size_out = self.voxel_size
        else:
            self.voxel_size_out = tuple(self.voxel_size_out)
        self.origin_file_name = file_name

        if (
            0 < len(self.projection_path)
            and self.projection_path[-1] != os.path.sep
        ):
            self.projection_path = self.projection_path + os.path.sep

        self.path_to_data = os.path.join(self.path_to_data, "")
        self.trsf_folder = os.path.join(self.trsf_folder, "")
        self.path_to_bin = os.path.join(self.path_to_bin, "")
        if 0 < len(self.path_to_bin) and not os.path.exists(self.path_to_bin):
            print("Binary path could not be found, will try with global call")
            self.path_to_bin = ""


class TimeRegistration:
    @staticmethod
    def read_trsf(path: str) -> np.ndarray:
        """
        Read a transformation from a text file

        Args:
            path (str): path to a transformation

        Returns
            (np.ndarray): 4x4 ndarray matrix
        """
        f = open(path)
        if f.read()[0] == "(":
            f.close()
            f = open(path)
            lines = f.readlines()[2:-1]
            f.close()
            return np.array([[float(v) for v in l.split()] for l in lines])
        f.close()
        return np.loadtxt(path)

    def __produce_trsf(self, params: Tuple[trsf_parameters, int, int, bool]):
        """
        Given a parameter object a reference time and two time points
        compute the transformation that registers together two consecutive in time images.

        Args:
            params (tuple): a tuple with the parameter object, two consecutive time points
                to register and a boolean to detemine wether to apply the trsf or not
        """
        (p, t1, t2, make) = params
        if not p.sequential:
            p_im_flo = p.A0.format(t=t2)
            p_im_ref = p.ref_path
            t_ref = p.ref_TP
            t_flo = t2
            if t1 == p.to_register[0]:
                self.__produce_trsf((p, t2, t1, make))
        else:
            if t1 < p.ref_TP:
                t_ref = t2
                t_flo = t1
            else:
                t_ref = t1
                t_flo = t2
            p_im_ref = p.A0.format(t=t_ref)
            p_im_flo = p.A0.format(t=t_flo)
        if not make:
            print("trsf tp %d-%d not done" % (t1, t2))
            np.savetxt(
                os.path.join(p.trsf_folder, "t%06d-%06d.txt" % (t_flo, t_ref)),
                np.identity(4),
            )
        elif t_flo != t_ref and (
            p.recompute
            or not os.path.exists(
                os.path.join(p.trsf_folder, "t%06d-%06d.txt" % (t_flo, t_ref))
            )
        ):
            if p.low_th is not None and 0 < p.low_th:
                th = " -ref-lt {lt:f} -flo-lt {lt:f} -no-norma ".format(
                    lt=p.low_th
                )
            else:
                th = ""
            if p.trsf_type != "vectorfield":
                if p.pre_2D == 1:
                    call(
                        self.path_to_bin
                        + "blockmatching -ref "
                        + p_im_ref
                        + " -flo "
                        + p_im_flo
                        + " -reference-voxel %f %f %f" % p.voxel_size
                        + " -floating-voxel %f %f %f" % p.voxel_size
                        + " -trsf-type rigid2D -py-hl %d -py-ll %d"
                        % (
                            p.registration_depth_start,
                            p.registration_depth_end,
                        )
                        + " -res-trsf "
                        + os.path.join(
                            p.trsf_folder,
                            "t%06d-%06d-tmp.txt" % (t_flo, t_ref),
                        )
                        + th,
                        shell=True,
                    )
                    pre_trsf = (
                        " -init-trsf "
                        + os.path.join(
                            p.trsf_folder,
                            "t%06d-%06d-tmp.txt" % (t_flo, t_ref),
                        )
                        + " -composition-with-initial "
                    )
                else:
                    pre_trsf = ""
                call(
                    self.path_to_bin
                    + "blockmatching -ref "
                    + p_im_ref
                    + " -flo "
                    + p_im_flo
                    + pre_trsf
                    + " -reference-voxel %f %f %f" % p.voxel_size
                    + " -floating-voxel %f %f %f" % p.voxel_size
                    + " -trsf-type %s -py-hl %d -py-ll %d"
                    % (
                        p.trsf_type,
                        p.registration_depth_start,
                        p.registration_depth_end,
                    )
                    + " -res-trsf "
                    + os.path.join(
                        p.trsf_folder, "t%06d-%06d.txt" % (t_flo, t_ref)
                    )
                    + th,
                    shell=True,
                )
            else:
                if p.apply_trsf:
                    res = " -res " + p.A0_out.format(t=t_flo)
                else:
                    res = ""
                if p.keep_vectorfield and pyklb_found:
                    res_trsf = (
                        " -composition-with-initial -res-trsf "
                        + os.path.join(
                            p.trsf_folder, "t%06d-%06d.klb" % (t_flo, t_ref)
                        )
                    )
                else:
                    res_trsf = ""
                    if p.keep_vectorfield:
                        print(
                            "The vectorfield cannot be stored without pyklb being installed"
                        )
                call(
                    self.path_to_bin
                    + "blockmatching -ref "
                    + p_im_ref
                    + " -flo "
                    + p_im_flo
                    + " -reference-voxel %f %f %f" % p.voxel_size
                    + " -floating-voxel %f %f %f" % p.voxel_size
                    + " -trsf-type affine -py-hl %d -py-ll %d"
                    % (p.registration_depth_start, p.registration_depth_end)
                    + " -res-trsf "
                    + os.path.join(
                        p.trsf_folder, "t%06d-%06d.txt" % (t_flo, t_ref)
                    )
                    + th,
                    shell=True,
                )
                call(
                    self.path_to_bin
                    + "blockmatching -ref "
                    + p_im_ref
                    + " -flo "
                    + p_im_flo
                    + " -init-trsf "
                    + os.path.join(
                        p.trsf_folder, "t%06d-%06d.txt" % (t_flo, t_ref)
                    )
                    + res
                    + " -reference-voxel %f %f %f" % p.voxel_size
                    + " -floating-voxel %f %f %f" % p.voxel_size
                    + " -trsf-type %s -py-hl %d -py-ll %d"
                    % (
                        p.trsf_type,
                        p.registration_depth_start,
                        p.registration_depth_end,
                    )
                    + res_trsf
                    + (
                        " -elastic-sigma {s:.1f} {s:.1f} {s:.1f} "
                        + " -fluid-sigma {s:.1f} {s:.1f} {s:.1f}"
                    ).format(s=p.sigma)
                    + th,
                    shell=True,
                )

    def run_produce_trsf(self, p: trsf_parameters):
        """
        Parallel processing of the transformations from t to t-1/t+1 to t (depending on t<r)
        The transformation is computed using blockmatching algorithm

        Args:
            p (trsf_paraneters): a trsf_parameter object
        """
        mapping = [
            (p, t1, t2, not (t1 in p.not_to_do or t2 in p.not_to_do))
            for t1, t2 in zip(p.to_register[:-1], p.to_register[1:])
        ]
        tic = time()
        tmp = []
        for mi in mapping:
            tmp += [self.__produce_trsf(mi)]
        tac = time()
        whole_time = tac - tic
        secs = whole_time % 60
        whole_time = whole_time // 60
        mins = whole_time % 60
        hours = whole_time // 60
        print("%dh:%dmin:%ds" % (hours, mins, secs))

    def compose_trsf(
        self, flo_t: int, ref_t: int, trsf_p: str, tp_list: List[int]
    ) -> str:
        """
        Recusrively build the transformation that allows
        to register time `flo_t` onto the frame of time `ref_t`
        assuming that it exists the necessary intermediary transformations

        Args:
            flo_t (int): time of the floating image
            ref_t (int): time of the reference image
            trsf_p (str): path to folder containing the transformations
            tp_list (list): list of time points that have been processed

        Returns:
            out_trsf (str): path to the result composed transformation
        """
        out_trsf = trsf_p + "t%06d-%06d.txt" % (flo_t, ref_t)
        if not os.path.exists(out_trsf):
            flo_int = tp_list[tp_list.index(flo_t) + np.sign(ref_t - flo_t)]
            # the call is recursive, to build `T_{flo\leftarrow ref}`
            # we need `T_{flo+1\leftarrow ref}` and `T_{flo\leftarrow ref-1}`
            trsf_1 = self.compose_trsf(flo_int, ref_t, trsf_p, tp_list)
            trsf_2 = self.compose_trsf(flo_t, flo_int, trsf_p, tp_list)
            call(
                self.path_to_bin
                + "composeTrsf "
                + out_trsf
                + " -trsfs "
                + trsf_2
                + " "
                + trsf_1,
                shell=True,
            )
        return out_trsf

    @staticmethod
    def __lowess_smooth(
        X: np.ndarray, T: np.ndarray, frac: float
    ) -> np.ndarray:
        """
        Smooth a curve using the lowess algorithm.
        See:
            Cleveland, W.S. (1979)
            “Robust Locally Weighted Regression and Smoothing Scatterplots”.
            Journal of the American Statistical Association 74 (368): 829-836.

        Args:
            X (np.ndarray): 1D array for the point positions
            T (np.ndarray): 1D array for the times corresponding to the `X` positions
            frac (float): Between 0 and 1. The fraction of the data used when estimating each value within X

        Returns:
            (np.ndarray): the smoothed X values
        """
        return lowess(X, T, frac=frac, is_sorted=True, return_sorted=False)

    @staticmethod
    def interpolation(
        p: trsf_parameters, X: np.ndarray, T: np.ndarray
    ) -> np.ndarray:
        """
        Interpolate positional values between missing timepoints using spline interplolation

        Args:
            p (trsf_parameters): parameters for the function
            X (np.ndarray): 1D array for the point positions
            T (np.ndarray): 1D array for the times corresponding to the `X` position

        Returns:
            (np.ndarray): The interoplated function
        """
        return sp.interpolate.InterpolatedUnivariateSpline(T, X, k=p.spline)

    @classmethod
    def read_param_file(clf, p_param: str = None) -> List[trsf_parameters]:
        """
        Asks for, reads and formats the parameter file

        Args:
            p_params (str | None): path to the json parameter files.
                It can either be a json or a folder containing json files.
                If it is None (default), a path is asked to the user.

        Returns:
            (list): list of `trsf_parameters` objects
        """
        if not isinstance(p_param, dict):
            if p_param is None:
                if len(sys.argv) < 2 or sys.argv[1] == "-f":
                    p_param = input(
                        "\nPlease inform the path to the json config file:\n"
                    )
                else:
                    p_param = sys.argv[1]
            stable = False or isinstance(p_param, Path)
            while not stable:
                tmp = p_param.strip('"').strip("'").strip(" ")
                stable = tmp == p_param
                p_param = tmp
            if os.path.isdir(p_param):
                f_names = [
                    os.path.join(p_param, f)
                    for f in os.listdir(p_param)
                    if ".json" in f and not "~" in f
                ]
            else:
                f_names = [p_param]
        else:
            f_names = [p_param]
        params = []
        for file_name in f_names:
            if isinstance(file_name, str):
                print("")
                print("Extraction of the parameters from file %s" % file_name)
            p = trsf_parameters(file_name)
            if not p.check_parameters_consistancy():
                print("\n%s Failed the consistancy check, it will be skipped")
            else:
                params += [p]
            print("")
        return params

    @staticmethod
    def prepare_paths(p: trsf_parameters):
        """
        Prepare the paths in a format that is usable by the algorithm

        Args:
            p (trsf_parameters): parameter object
        """
        image_formats = [
            ".tiff",
            ".tif",
            ".inr",
            ".gz",
            ".klb",
            ".h5",
            ".hdf5",
        ]
        ### Check the file names and folders:
        p.im_ext = p.file_name.split(".")[-1]  # Image type
        p.A0 = os.path.join(p.path_to_data, p.file_name)  # Image path
        # Output format
        if p.output_format is not None:
            # If the output format is a file alone
            if os.path.split(p.output_format)[0] == "":
                p.A0_out = os.path.join(p.path_to_data, p.output_format)
            # If the output format is a folder alone
            elif not os.path.splitext(p.output_format)[-1] in image_formats:
                p.A0_out = os.path.join(p.output_format, p.file_name)
            else:
                p.A0_out = p.output_format
        else:
            p.A0_out = os.path.join(
                p.path_to_data,
                p.file_name.replace(p.im_ext, p.suffix + "." + p.im_ext),
            )

        # Time points to work with
        p.time_points = np.array(
            [i for i in np.arange(p.first, p.last + 1) if not i in p.not_to_do]
        )
        if p.check_TP:
            missing_time_points = []
            for t in p.time_points:
                if not os.path.exists(p.A0.format(t=t)):
                    missing_time_points += [t]
            if len(missing_time_points) != 0:
                print("The following time points are missing:")
                print("\t" + missing_time_points)
                print("Aborting the process")
                exit()
        if not p.sequential:
            p.ref_path = p.A0.format(t=p.ref_TP)
        if p.apply_trsf:
            for t in sorted(p.time_points):
                folder_tmp = os.path.split(p.A0_out.format(t=t))[0]
                if not os.path.exists(folder_tmp):
                    os.makedirs(folder_tmp)

        max_t = max(p.time_points)
        if p.trsf_interpolation:
            p.to_register = sorted(p.time_points)[:: p.step_size]
            if not max_t in p.to_register:
                p.to_register += [max_t]
        else:
            p.to_register = p.time_points
        if not hasattr(p, "bdv_im") or p.bdv_im is None:
            p.bdv_im = p.A0
        if not hasattr(p, "out_bdv") or p.out_bdv is None:
            p.out_bdv = os.path.join(p.trsf_folder, "bdv.xml")
        if p.bdv_voxel_size is None:
            p.bdv_voxel_size = p.voxel_size

    def lowess_filter(self, p: trsf_parameters, trsf_fmt: str) -> str:
        """
        Apply a lowess filter on a set of ordered transformations (only works for translations).
        It does it from already computed transformations and write new transformations on disk.

        Args:
            p (trsf_parameters): parameter object
            trsf_fmt (srt): the string format for the initial transformations

        Returns:
            (str): output transformation string
        """
        X_T = []
        Y_T = []
        Z_T = []
        T = []
        for t in p.to_register:
            trsf_p = p.trsf_folder + trsf_fmt.format(flo=t, ref=p.ref_TP)
            trsf = self.read_trsf(trsf_p)
            T += [t]
            X_T += [trsf[0, -1]]
            Y_T += [trsf[1, -1]]
            Z_T += [trsf[2, -1]]

        frac = float(p.window_size) / len(T)
        X_smoothed = self.__lowess_smooth(p, X_T, T, frac=frac)
        Y_smoothed = self.__lowess_smooth(p, Y_T, T, frac=frac)
        Z_smoothed = self.__lowess_smooth(p, Z_T, T, frac=frac)

        new_trsf_fmt = "t{flo:06d}-{ref:06d}-filtered.txt"
        for i, t in enumerate(T):
            mat = np.identity(4)
            mat[0, -1] = X_smoothed[i]
            mat[1, -1] = Y_smoothed[i]
            mat[2, -1] = Z_smoothed[i]
            np.savetxt(
                p.trsf_folder + new_trsf_fmt.format(flo=t, ref=p.ref_TP), mat
            )
        return new_trsf_fmt

    def interpolate(
        self, p: trsf_parameters, trsf_fmt: str, new_trsf_fmt: str = None
    ) -> str:
        """
        Interpolate a set of ordered transformations (only works for translations).
        It does it from already computed transformations and write new transformations on disk.

        Args:
            p (trsf_parameters): parameter object
            trsf_fmt (srt): the string format for the initial transformations
            new_trsf_fmt (str | None): path to the interpolated transformations. If None,
                the format will be "t{flo:06d}-{ref:06d}-interpolated.txt"

        Returns:
            (str): output transformation format
        """
        if new_trsf_fmt is None:
            new_trsf_fmt = "t{flo:06d}-{ref:06d}-interpolated.txt"
        X_T = []
        Y_T = []
        Z_T = []
        T = []
        for t in p.to_register:
            trsf_p = os.path.join(
                p.trsf_folder, trsf_fmt.format(flo=t, ref=p.ref_TP)
            )
            trsf = self.read_trsf(trsf_p)
            T += [t]
            X_T += [trsf[0, -1]]
            Y_T += [trsf[1, -1]]
            Z_T += [trsf[2, -1]]

        X_interp = self.interpolation(p, X_T, T)
        Y_interp = self.interpolation(p, Y_T, T)
        Z_interp = self.interpolation(p, Z_T, T)
        for t in p.time_points:
            mat = np.identity(4)
            mat[0, -1] = X_interp(t)
            mat[1, -1] = Y_interp(t)
            mat[2, -1] = Z_interp(t)
            np.savetxt(
                os.path.join(
                    p.trsf_folder,
                    new_trsf_fmt.format(flo=t, ref=p.ref_TP),
                    mat,
                )
            )
        trsf_fmt = new_trsf_fmt
        return trsf_fmt

    @staticmethod
    def pad_trsfs(p: trsf_parameters, trsf_fmt: str):
        """
        Pad transformations

        Args:
            p (trsf_parameters): parameter object
            trsf_fmt (srt): the string format for the initial transformations
        """
        if not p.sequential and p.ref_path.split(".")[-1] == "klb":
            im_shape = readheader(p.ref_path)["imagesize_tczyx"][-1:-4:-1]
        elif not p.sequential:
            im_shape = imread(p.ref_path).shape
        elif p.A0.split(".")[-1] == "klb":
            im_shape = readheader(p.A0.format(t=p.ref_TP))["imagesize_tczyx"][
                -1:-4:-1
            ]
        else:
            im_shape = imread(p.A0.format(t=p.ref_TP)).shape
        im = SpatialImage(np.ones(im_shape), dtype=np.uint8)
        im.voxelsize = p.voxel_size
        if pyklb_found:
            template = os.path.join(p.trsf_folder, "tmp.klb")
            res_t = "template.klb"
        else:
            template = os.path.join(p.trsf_folder, "tmp.tif")
            res_t = "template.tif"

        imsave(template, im)
        identity = np.identity(4)

        trsf_fmt_no_flo = trsf_fmt.replace("{flo:06d}", "%06d")
        new_trsf_fmt = "t{flo:06d}-{ref:06d}-padded.txt"
        new_trsf_fmt_no_flo = new_trsf_fmt.replace("{flo:06d}", "%06d")
        for t in p.not_to_do:
            np.savetxt(
                os.path.join(
                    p.trsf_folder,
                    trsf_fmt.format(flo=t, ref=p.ref_TP),
                    identity,
                )
            )

        call(
            p.path_to_bin
            + "changeMultipleTrsfs -trsf-format "
            + os.path.join(p.trsf_folder, trsf_fmt_no_flo.format(ref=p.ref_TP))
            + " -index-reference %d -first %d -last %d "
            % (p.ref_TP, min(p.time_points), max(p.time_points))
            + " -template "
            + template
            + " -res "
            + os.path.join(
                p.trsf_folder, new_trsf_fmt_no_flo.format(ref=p.ref_TP)
            )
            + " -res-t "
            + os.path.join(p.trsf_folder, res_t)
            + " -trsf-type %s -vs %f %f %f" % ((p.trsf_type,) + p.voxel_size),
            shell=True,
        )

    def compute_trsfs(self, p: trsf_parameters):
        """
        Compute all the transformations from a given set of parameters

        Args:
            p (trsf_parameters): Parameter object
        """
        # Create the output folder for the transfomrations
        if not os.path.exists(p.trsf_folder):
            os.makedirs(p.trsf_folder)

        trsf_fmt = "t{flo:06d}-{ref:06d}.txt"
        try:
            self.run_produce_trsf(p)
            if p.sequential:
                if min(p.to_register) != p.ref_TP:
                    self.compose_trsf(
                        min(p.to_register),
                        p.ref_TP,
                        p.trsf_folder,
                        list(p.to_register),
                    )
                if max(p.to_register) != p.ref_TP:
                    self.compose_trsf(
                        max(p.to_register),
                        p.ref_TP,
                        p.trsf_folder,
                        list(p.to_register),
                    )
            np.savetxt(
                os.path.join("{:s}", trsf_fmt).format(
                    p.trsf_folder, flo=p.ref_TP, ref=p.ref_TP
                ),
                np.identity(4),
            )
        except Exception as e:
            print(p.trsf_folder)
            print(e)

        if p.lowess:
            trsf_fmt = self.lowess_filter(p, trsf_fmt)
        if p.trsf_interpolation:
            trsf_fmt = interpolate(p, trsf_fmt)
        if p.padding:
            self.pad_trsfs(p, trsf_fmt)

    @staticmethod
    def apply_trsf(p: trsf_parameters):
        """
        Apply transformations to a movie from a set of computed transformations

        Args:
            p (trsf_parameters): Parameter object
        """
        trsf_fmt = "t{flo:06d}-{ref:06d}.txt"
        if p.lowess:
            trsf_fmt = "t{flo:06d}-{ref:06d}-filtered.txt"
        if p.trsf_interpolation:
            trsf_fmt = "t{flo:06d}-{ref:06d}-interpolated.txt"
        if p.padding:
            trsf_fmt = "t{flo:06d}-{ref:06d}-padded.txt"
            if pyklb_found:
                template = os.path.join(p.trsf_folder, "template.klb")
                X, Y, Z = readheader(template)["imagesize_tczyx"][-1:-4:-1]
            else:
                template = os.path.join(p.trsf_folder, "template.tif")
                X, Y, Z = imread(template).shape
        elif p.A0.split(".")[-1] == "klb":
            X, Y, Z = readheader(p.A0.format(t=p.ref_TP))["imagesize_tczyx"][
                -1:-4:-1
            ]
            template = p.A0.format(t=p.ref_TP)
        else:
            X, Y, Z = imread(p.A0.format(t=p.ref_TP)).shape
            template = p.A0.format(t=p.ref_TP)

        if p.voxel_size != p.voxel_size_out:
            before, after = os.path.splitext(template)
            old_template = template
            template = "".join((before, ".final_template", after))
            call(
                p.path_to_bin
                + f"applyTrsf {old_template} {template} "
                + "-vs %f %f %f" % p.voxel_size_out,
                shell=True,
            )
            X, Y, Z = imread(template).shape

        xy_proj = np.zeros((X, Y, len(p.time_points)), dtype=np.uint16)
        xz_proj = np.zeros((X, Z, len(p.time_points)), dtype=np.uint16)
        yz_proj = np.zeros((Y, Z, len(p.time_points)), dtype=np.uint16)
        for i, t in enumerate(sorted(p.time_points)):
            folder_tmp = os.path.split(p.A0_out.format(t=t))[0]
            if not os.path.exists(folder_tmp):
                os.makedirs(folder_tmp)
            call(
                p.path_to_bin
                + "applyTrsf %s %s -trsf "
                % (p.A0.format(t=t), p.A0_out.format(t=t))
                + os.path.join(
                    p.trsf_folder, trsf_fmt.format(flo=t, ref=p.ref_TP)
                )
                + " -template "
                + template
                + " -floating-voxel %f %f %f " % p.voxel_size
                + " -reference-voxel %f %f %f " % p.voxel_size_out
                + " -interpolation %s" % p.image_interpolation,
                shell=True,
            )
            try:
                im = imread(p.A0_out.format(t=t))
            except Exception as e:
                # print("applyTrsf failed at t=",str(t),", retrying now")
                call(
                    p.path_to_bin
                    + "applyTrsf %s %s -trsf "
                    % (p.A0.format(t=t), p.A0_out.format(t=t))
                    + os.path.join(
                        p.trsf_folder, trsf_fmt.format(flo=t, ref=p.ref_TP)
                    )
                    + " -template "
                    + template
                    + " -floating-voxel %f %f %f " % p.voxel_size
                    + " -reference-voxel %f %f %f " % p.voxel_size_out
                    + " -interpolation %s" % p.image_interpolation,
                    shell=True,
                )
                im = imread(p.A0_out.format(t=t))
            if p.projection_path is not None:
                xy_proj[..., i] = SpatialImage(np.max(im, axis=2))
                xz_proj[..., i] = SpatialImage(np.max(im, axis=1))
                yz_proj[..., i] = SpatialImage(np.max(im, axis=0))
        if p.projection_path is not None:
            if not os.path.exists(p.projection_path):
                os.makedirs(p.projection_path)
            p_to_data = p.projection_path
            num_s = p.file_name.find("{")
            num_e = p.file_name.find("}") + 1
            f_name = p.file_name.replace(p.file_name[num_s:num_e], "")
            if not os.path.exists(p_to_data.format(t=-1)):
                os.makedirs(p_to_data.format(t=-1))
            imsave(
                os.path.join(
                    p_to_data, f_name.replace(p.im_ext, "xyProjection.tif")
                ),
                SpatialImage(xy_proj),
            )
            imsave(
                os.path.join(
                    p_to_data, f_name.replace(p.im_ext, "xzProjection.tif")
                ),
                SpatialImage(xz_proj),
            )
            imsave(
                os.path.join(
                    p_to_data, f_name.replace(p.im_ext, "yzProjection.tif")
                ),
                SpatialImage(yz_proj),
            )

    @staticmethod
    def inv_trsf(trsf: np.ndarray) -> np.ndarray:
        """
        Inverse a transformation

        Args:
            trsf (np.ndarray): 4x4 transformation matrix

        Returns
            (np.ndarray): the inverse of the input transformation
        """
        return np.linalg.lstsq(trsf, np.identity(4))[0]

    @staticmethod
    def prettify(elem: ET.Element) -> str:
        """
        Return a pretty-printed XML string for the Element.

        Args:
            elem (xml.etree.ElementTree.Element): xml element

        Returns:
            (str): a nice version of our xml file
        """
        rough_string = ET.tostring(elem, "utf-8")
        reparsed = minidom.parseString(rough_string)
        return reparsed.toprettyxml(indent="  ")

    @staticmethod
    def do_viewSetup(
        ViewSetup: ET.SubElement,
        p: trsf_parameters,
        im_size: Tuple[int, int, int],
        i: int,
    ):
        """
        Setup xml elements for BigDataViewer

        Args:
            ViewSetup (ET.SubElement): ...
            p (trsf_parameters): Parameter object
            im_size (tuple): tuple of 3 integers with the dimension of the image
            i (int): angle id
        """
        id_ = ET.SubElement(ViewSetup, "id")
        id_.text = "%d" % i
        name = ET.SubElement(ViewSetup, "name")
        name.text = "%d" % i
        size = ET.SubElement(ViewSetup, "size")
        size.text = "%d %d %d" % tuple(im_size)
        voxelSize = ET.SubElement(ViewSetup, "voxelSize")
        unit = ET.SubElement(voxelSize, "unit")
        unit.text = p.bdv_unit
        size = ET.SubElement(voxelSize, "size")
        size.text = "%f %f %f" % tuple(p.bdv_voxel_size)
        attributes = ET.SubElement(ViewSetup, "attributes")
        illumination = ET.SubElement(attributes, "illumination")
        illumination.text = "0"
        channel = ET.SubElement(attributes, "channel")
        channel.text = "0"
        tile = ET.SubElement(attributes, "tile")
        tile.text = "0"
        angle = ET.SubElement(attributes, "angle")
        angle.text = "%d" % i

    def do_ViewRegistration(
        self,
        ViewRegistrations: ET.SubElement,
        p: trsf_parameters,
        t: int,
        a: int,
    ):
        """
        Write the view registration for BigDataViewer

        Args:
            ViewSetup (ET.SubElement): ...
            p (trsf_parameters): Parameter object
            t (int): time point to treat
            a (int): angle to treat (useless for time regisrtation)
        """
        ViewRegistration = ET.SubElement(ViewRegistrations, "ViewRegistration")
        ViewRegistration.set("timepoint", "%d" % t)
        ViewRegistration.set("setup", "0")
        ViewTransform = ET.SubElement(ViewRegistration, "ViewTransform")
        ViewTransform.set("type", "affine")
        affine = ET.SubElement(ViewTransform, "affine")
        f = os.path.join(p.trsf_folder, "t%06d-%06d.txt" % (t, p.ref_TP))
        trsf = self.read_trsf(f)
        trsf = self.inv_trsf(trsf)
        formated_trsf = tuple(trsf[:-1, :].flatten())
        affine.text = ("%f " * 12) % formated_trsf
        ViewTransform = ET.SubElement(ViewRegistration, "ViewTransform")
        ViewTransform.set("type", "affine")
        affine = ET.SubElement(ViewTransform, "affine")
        affine.text = (
            "%f 0.0 0.0 0.0 0.0 %f 0.0 0.0 0.0 0.0 %f 0.0" % p.voxel_size
        )

    def build_bdv(self, p: trsf_parameters):
        """
        Build the BigDataViewer xml

        Args:
            p (trsf_parameters): Parameter object
        """
        if not p.im_ext in ["klb"]:  # ['tif', 'klb', 'tiff']:
            print("Image format not adapted for BigDataViewer")
            return
        SpimData = ET.Element("SpimData")
        SpimData.set("version", "0.2")
        SpimData.set("encoding", "UTF-8")

        base_path = ET.SubElement(SpimData, "BasePath")
        base_path.set("type", "relative")
        base_path.text = "."

        SequenceDescription = ET.SubElement(SpimData, "SequenceDescription")

        ImageLoader = ET.SubElement(SequenceDescription, "ImageLoader")
        ImageLoader.set("format", p.im_ext)
        Resolver = ET.SubElement(ImageLoader, "Resolver")
        Resolver.set(
            "type", "org.janelia.simview.klb.bdv.KlbPartitionResolver"
        )
        ViewSetupTemplate = ET.SubElement(Resolver, "ViewSetupTemplate")
        template = ET.SubElement(ViewSetupTemplate, "template")
        template.text = p.bdv_im.format(t=p.to_register[0])
        timeTag = ET.SubElement(ViewSetupTemplate, "timeTag")
        timeTag.text = p.time_tag

        ViewSetups = ET.SubElement(SequenceDescription, "ViewSetups")
        ViewSetup = ET.SubElement(ViewSetups, "ViewSetup")
        if p.im_ext == "klb":
            im_size = tuple(
                readheader(p.A0.format(t=p.ref_TP))["imagesize_tczyx"][
                    -1:-4:-1
                ]
            )
        else:
            im_size = tuple(imread(p.A0.format(t=p.ref_TP)).shape)
        self.do_viewSetup(ViewSetup, p, im_size, 0)

        Attributes = ET.SubElement(ViewSetups, "Attributes")
        Attributes.set("name", "illumination")
        Illumination = ET.SubElement(Attributes, "Illumination")
        id_ = ET.SubElement(Illumination, "id")
        id_.text = "0"
        name = ET.SubElement(Illumination, "name")
        name.text = "0"

        Attributes = ET.SubElement(ViewSetups, "Attributes")
        Attributes.set("name", "channel")
        Channel = ET.SubElement(Attributes, "Channel")
        id_ = ET.SubElement(Channel, "id")
        id_.text = "0"
        name = ET.SubElement(Channel, "name")
        name.text = "0"

        Attributes = ET.SubElement(ViewSetups, "Attributes")
        Attributes.set("name", "tile")
        Tile = ET.SubElement(Attributes, "Tile")
        id_ = ET.SubElement(Tile, "id")
        id_.text = "0"
        name = ET.SubElement(Tile, "name")
        name.text = "0"

        Attributes = ET.SubElement(ViewSetups, "Attributes")
        Attributes.set("name", "angle")
        Angle = ET.SubElement(Attributes, "Angle")
        id_ = ET.SubElement(Angle, "id")
        id_.text = "0"
        name = ET.SubElement(Angle, "name")
        name.text = "0"

        TimePoints = ET.SubElement(SequenceDescription, "Timepoints")
        TimePoints.set("type", "range")
        first = ET.SubElement(TimePoints, "first")
        first.text = "%d" % min(p.to_register)
        last = ET.SubElement(TimePoints, "last")
        last.text = "%d" % max(p.to_register)
        ViewRegistrations = ET.SubElement(SpimData, "ViewRegistrations")
        b = min(p.to_register)
        e = max(p.to_register)
        for t in range(b, e + 1):
            self.do_ViewRegistration(ViewRegistrations, p, t)

        with open(p.out_bdv, "w") as f:
            f.write(self.prettify(SpimData))
            f.close()

    def plot_transformations(self, p: trsf_parameters):
        trsf_fmt = "t{flo:06d}-{ref:06d}.txt"
        if p.lowess:
            trsf_fmt = "t{flo:06d}-{ref:06d}-filtered.txt"
        if p.trsf_interpolation:
            trsf_fmt = "t{flo:06d}-{ref:06d}-interpolated.txt"
        if p.padding:
            trsf_fmt = "t{flo:06d}-{ref:06d}-padded.txt"

        import matplotlib.pyplot as plt

        tX, tY, tZ = [], [], []
        rX, rY, rZ = [], [], []
        for t in sorted(p.time_points):
            trsf = self.read_trsf(
                os.path.join(
                    p.trsf_folder, trsf_fmt.format(flo=t, ref=p.ref_TP)
                )
            )
            (tx, ty, tz), M, *_ = decompose(trsf)
            rx, ry, rz = mat2euler(M)
            tX.append(tx)
            tY.append(ty)
            tZ.append(tz)
            rX.append(np.rad2deg(rx))
            rY.append(np.rad2deg(ry))
            rZ.append(np.rad2deg(rz))
        fig, ax = plt.subplots(3, 1, figsize=(8, 5), sharex=True, sharey=True)
        ax[0].plot(p.time_points, tX, "o-")
        ax[1].plot(p.time_points, tY, "o-")
        ax[2].plot(p.time_points, tZ, "o-")
        for axis, axi in zip(["X", "Y", "Z"], ax):
            axi.set_ylabel(f"{axis} Translation [µm]")
        ax[2].set_xlabel("Time")
        fig.suptitle("Translations")
        fig.tight_layout()

        fig, ax = plt.subplots(3, 1, figsize=(8, 5), sharex=True, sharey=True)
        ax[0].plot(p.time_points, rX, "o-")
        ax[1].plot(p.time_points, rY, "o-")
        ax[2].plot(p.time_points, rZ, "o-")
        for axis, axi in zip(["X", "Y", "Z"], ax):
            axi.set_ylabel(f"{axis} Rotation\nin degree")
        ax[2].set_xlabel("Time")
        fig.suptitle("Rotations")
        fig.tight_layout()

        plt.show()

    def run_trsf(self):
        """
        Start the Spatial registration after having informed the parameter files
        """
        for p in self.params:
            try:
                print("Starting experiment")
                print(p)
                self.prepare_paths(p)
                if p.compute_trsf:
                    self.compute_trsfs(p)
                if p.plot_trsf:
                    self.plot_transformations(p)
                if p.apply_trsf and p.trsf_type != "vectorfield":
                    self.apply_trsf(p)
                if p.do_bdv:
                    self.build_bdv(p)
            except Exception as e:
                print("Failure of %s" % p.origin_file_name)
                print(e)

    def __init__(self, params=None):
        if params is None:
            self.params = self.read_param_file()
        elif (
            isinstance(params, str)
            or isinstance(params, Path)
            or isinstance(params, dict)
        ):
            self.params = TimeRegistration.read_param_file(params)
        else:
            self.params = params
        if self.params is not None and 0 < len(self.params):
            self.path_to_bin = self.params[0].path_to_bin


def time_registration():
    reg = TimeRegistration()
    reg.run_trsf()



#!/usr/bin/python
# This file is subject to the terms and conditions defined in
# file 'LICENCE', which is part of this source code package.
# Author: Leo Guignard (leo.guignard...@AT@...univ-amu.fr)

import numpy as np
import os
import sys
import json
from subprocess import call
import xml.etree.ElementTree as ET
from xml.dom import minidom
from shutil import copyfile
from pathlib import Path
from typing import Union, List, Tuple
from IO import imread, imsave, SpatialImage


class trsf_parameters(object):
    """
    Read parameters for the registration function from a preformated json file
    """

    def check_parameters_consistancy(self) -> bool:
        """
        Function that should check parameter consistancy

        @TODO:
        write something that actually do check
        """
        correct = True
        if not (
            hasattr(self, "out_pattern") or hasattr(self, "output_format")
        ):
            print("The output pattern cannot be an empty string")
            correct = False
        return correct

    def __str__(self) -> str:
        max_key = (
            max([len(k) for k in self.__dict__.keys() if k != "param_dict"])
            + 1
        )
        output = "The registration will run with the following arguments:\n"
        output += "\n" + " File format \n"
        output += "path_to_data".ljust(max_key, " ") + ": {:s}\n".format(
            self.path_to_data
        )
        output += "ref_im".ljust(max_key, " ") + ": {:s}\n".format(self.ref_im)

        output += "flo_ims".ljust(max_key, " ") + ": "
        tmp_just_len = len("flo_ims".ljust(max_key, " ") + ": ")
        already = tmp_just_len + 1
        for flo in self.flo_ims:
            output += (" " * (tmp_just_len - already)) + "{:s}\n".format(flo)
            already = 0

        output += "init_trsfs".ljust(max_key, " ") + ": "
        tmp_just_len = len("init_trsfs".ljust(max_key, " ") + ": ")
        already = tmp_just_len + 1
        for init_trsfs in self.init_trsfs:
            if init_trsfs is not None and 0 < len(init_trsfs):
                output += (" " * (tmp_just_len - already)) + "{}\n".format(
                    init_trsfs
                )
            already = 0

        output += "trsf_types".ljust(max_key, " ") + ": "
        tmp_just_len = len("trsf_types".ljust(max_key, " ") + ": ")
        already = tmp_just_len + 1
        for trsf_type in self.trsf_types:
            output += (" " * (tmp_just_len - already)) + "{:s}\n".format(
                trsf_type
            )
            already = 0

        output += "ref_voxel".ljust(
            max_key, " "
        ) + ": {:f} x {:f} x {:f}\n".format(*self.ref_voxel)

        output += "flo_voxels".ljust(max_key, " ") + ": "
        tmp_just_len = len("flo_voxels".ljust(max_key, " ") + ": ")
        already = tmp_just_len + 1
        for flo_voxel in self.flo_voxels:
            output += (
                " " * (tmp_just_len - already)
            ) + "{:f} x {:f} x {:f}\n".format(*flo_voxel)
            already = 0

        output += "out_voxel".ljust(
            max_key, " "
        ) + ": {:f} x {:f} x {:f}\n".format(*self.out_voxel)

        output += "out_pattern".ljust(max_key, " ") + ": {:s}\n".format(
            self.out_pattern
        )

        return output

    def add_path_prefix(self, prefix: str):
        """
        Add a prefix to all the relevant paths (this is mainly for the unitary tests)

        Args:
            prefix (str): The prefix to add in front of the path
        """
        self.path_to_data = str(Path(prefix) / self.path_to_data)
        self.trsf_paths = [
            str(Path(prefix) / trsf) for trsf in self.trsf_paths
        ]

    def __init__(self, file_name: str):
        if not isinstance(file_name, dict):
            with open(file_name) as f:
                param_dict = json.load(f)
                f.close()
        else:
            param_dict = {}
            for k, v in file_name.items():
                if isinstance(v, Path):
                    param_dict[k] = str(v)
                else:
                    param_dict[k] = v

        # Default parameters
        self.param_dict = param_dict
        self.init_trsfs = [[], [], []]
        self.path_to_bin = ""
        self.registration_depth_start = 6
        self.registration_depth_end = 3
        self.init_trsf_real_unit = True
        self.image_interpolation = "linear"
        self.apply_trsf = True
        self.compute_trsf = True
        self.test_init = False
        self.begin = None
        self.end = None
        self.trsf_types = []
        self.time_tag = None
        self.bdv_unit = "microns"
        self.bdv_voxel_size = None
        self.do_bdv = 0
        self.flo_im_sizes = None
        self.copy_ref = False
        self.bbox_out = False

        if "registration_depth" in param_dict:
            self.__dict__["registration_depth_start"] = 6
            self.__dict__["registration_depth_end"] = param_dict[
                "registration_depth"
            ]
        elif not "registration_depth_start" in param_dict:
            self.__dict__["registration_depth_start"] = 6
            self.__dict__["registration_depth_end"] = 3

        self.__dict__.update(param_dict)
        self.ref_voxel = tuple(self.ref_voxel)
        self.flo_voxels = [tuple(vox) for vox in self.flo_voxels]
        self.out_voxel = tuple(self.out_voxel)
        self.origin_file_name = file_name
        self.path_to_bin = os.path.join(self.path_to_bin, "")
        if 0 < len(self.path_to_bin) and not os.path.exists(self.path_to_bin):
            print("Binary path could not be found, will try with global call")
            self.path_to_bin = ""


class SpatialRegistration:
    @staticmethod
    def axis_rotation_matrix(
        axis: str,
        angle: float,
        min_space: Tuple[int, int, int] = None,
        max_space: Tuple[int, int, int] = None,
    ) -> np.ndarray:
        """Return the transformation matrix from the axis and angle necessary

        Args:
            axis (str): axis of rotation ("X", "Y" or "Z")
            angle (float) : angle of rotation (in degree)
            min_space (tuple(int, int, int)): coordinates of the bottom point (usually (0, 0, 0))
            max_space (tuple(int, int, int)): coordinates of the top point (usually im shape)

        Returns:
            (ndarray): 4x4 rotation matrix
        """
        import math

        I = np.linalg.inv
        D = np.dot
        if axis not in ["X", "Y", "Z"]:
            raise Exception(f"Unknown axis: {axis}")
        rads = math.radians(angle)
        s = math.sin(rads)
        c = math.cos(rads)

        centering = np.identity(4)
        if min_space is None and max_space is not None:
            min_space = np.array([0.0, 0.0, 0.0])

        if max_space is not None:
            space_center = (max_space - min_space) / 2.0
            offset = -1.0 * space_center
            centering[:3, 3] = offset

        rot = np.identity(4)
        if axis == "X":
            rot = np.array(
                [
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, c, -s, 0.0],
                    [0.0, s, c, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )
        elif axis == "Y":
            rot = np.array(
                [
                    [c, 0.0, s, 0.0],
                    [0.0, 1.0, 0.0, 0.0],
                    [-s, 0.0, c, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )

        elif axis == "Z":
            rot = np.array(
                [
                    [c, -s, 0.0, 0.0],
                    [s, c, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )

        return D(I(centering), D(rot, centering))

    @staticmethod
    def flip_matrix(axis: str, im_size: tuple) -> np.ndarray:
        """
        Build a matrix to flip an image according to a given axis

        Args:
            axis (str): axis along the flip is done ("X", "Y" or "Z")
            im_size (tuple(int, int, int)): coordinates of the top point (usually im shape)

        Returns:
            (ndarray): 4x4 flipping matrix
        """
        out = np.identity(4)
        if axis == "X":
            out[0, 0] = -1
            out[0, -1] = im_size[0]
        if axis == "Y":
            out[1, 1] = -1
            out[1, -1] = im_size[1]
        if axis == "Z":
            out[2, 2] = -1
            out[2, -1] = im_size[2]
        return out

    @staticmethod
    def translation_matrix(axis: str, tr: float) -> np.ndarray:
        """
        Build a matrix to flip an image according to a given axis

        Args:
            axis (str): axis along the flip is done ("X", "Y" or "Z")
            tr (float): translation value

        Returns:
            (ndarray): 4x4 flipping matrix
        """
        out = np.identity(4)
        if axis == "X":
            out[0, -1] = tr
        if axis == "Y":
            out[1, -1] = tr
        if axis == "Z":
            out[2, -1] = tr
        return out

    @classmethod
    def read_param_file(
        clf, p_param: trsf_parameters = None
    ) -> trsf_parameters:
        """
        Asks for, reads and formats the parameter file
        """
        if not isinstance(p_param, dict):
            if p_param is None:
                if len(sys.argv) < 2 or sys.argv[1] == "-f":
                    p_param = input(
                        "\nPlease inform the path to the json config file:\n"
                    )
                else:
                    p_param = sys.argv[1]
            stable = False or isinstance(p_param, Path)
            while not stable:
                tmp = p_param.strip('"').strip("'").strip(" ")
                stable = tmp == p_param
                p_param = tmp
            if os.path.isdir(p_param):
                f_names = [
                    os.path.join(p_param, f)
                    for f in os.listdir(p_param)
                    if ".json" in f and not "~" in f
                ]
            else:
                f_names = [p_param]
        else:
            f_names = [p_param]
        params = []
        for file_name in f_names:
            if isinstance(file_name, str):
                print("")
                print("Extraction of the parameters from file %s" % file_name)
            p = trsf_parameters(file_name)
            if not p.check_parameters_consistancy():
                print("\n%s Failed the consistancy check, it will be skipped")
            else:
                params += [p]
            print("")
        return params

    @staticmethod
    def read_trsf(path: Union[str, Path]) -> np.ndarray:
        """
        Read a transformation from a text file

        Args:
            path (str | Path): path to a transformation

        Returns:
            (ndarray): 4x4 transformation matrix
        """
        f = open(path)
        if f.read()[0] == "(":
            f.close()
            f = open(path)
            lines = f.readlines()[2:-1]
            f.close()
            return np.array([[float(v) for v in l.split()] for l in lines])
        else:
            f.close()
            return np.loadtxt(path)

    @staticmethod
    def prepare_paths(p: trsf_parameters):
        """
        Prepare the paths in a format that is usable by the algorithm

        Args:
            p (trsf_parameters): path to a transformation
        """
        p.ref_A = os.path.join(p.path_to_data, p.ref_im)
        p.flo_As = []
        for flo_im in p.flo_ims:
            p.flo_As += [os.path.join(p.path_to_data, flo_im)]
        if os.path.split(p.out_pattern)[0] == "":
            ext = p.ref_im.split(".")[-1]
            p.ref_out = p.ref_A.replace(ext, p.out_pattern + "." + ext)
            p.flo_outs = []
            for flo in p.flo_As:
                p.flo_outs += [flo.replace(ext, p.out_pattern + "." + ext)]
        else:
            if not os.path.exists(p.out_pattern):
                os.makedirs(p.out_pattern)
            p.ref_out = os.path.join(p.out_pattern, p.ref_im)
            p.flo_outs = []
            for flo in p.flo_ims:
                p.flo_outs += [os.path.join(p.out_pattern, flo)]
        if not hasattr(p, "trsf_paths"):
            p.trsf_paths = [os.path.split(pi)[0] for pi in p.flo_outs]
            p.trsf_names = ["A{a:d}-{trsf:s}.trsf" for _ in p.flo_outs]
        else:
            formated_paths = []
            p.trsf_names = []
            for pi in p.trsf_paths:
                path, n = os.path.split(pi)
                if os.path.splitext(n)[-1] == "":
                    n = ""
                    path = pi
                    if not os.path.exists(path):
                        os.makedirs(path)
                if n == "":
                    n = "A{a:d}-{trsf:s}.trsf"
                elif not "{a:" in n:
                    if not "{trsf:" in n:
                        n += "{a:d}-{trsf:s}.trsf"
                    else:
                        n += "{a:d}.trsf"
                elif not "{trsf:" in n:
                    n += "{trsf:s}.trsf"
                formated_paths += [path]
                p.trsf_names += [n]
            p.trsf_paths = formated_paths
        if p.time_tag is not None:
            s = p.ref_A.find(p.time_tag) + len(p.time_tag)
            e = s
            while p.ref_A[e].isdigit() and e < len(p.ref_A):
                e += 1
            p.begin = p.end = int(p.ref_A[s:e])
        else:
            p.begin = p.end = None
        if p.flo_im_sizes is None:
            p.flo_im_sizes = []
            for im_p in p.flo_As:
                p.flo_im_sizes.append(imread(im_p).shape)
        if (
            not hasattr(p, "ref_im_size") or p.ref_im_size is None
        ) and p.flo_im_sizes is not None:
            p.ref_im_size = p.flo_im_sizes[0]
        else:
            p.ref_im_size = imread(p.ref_A).shape
        if not hasattr(p, "bdv_im") or p.bdv_im is None:
            p.bdv_im = [p.ref_A] + p.flo_As
        if not hasattr(p, "out_bdv") or p.out_bdv is None:
            p.out_bdv = os.path.join(p.trsf_paths[0], "bdv.xml")
        if p.bdv_voxel_size is None:
            p.bdv_voxel_size = p.ref_voxel

    @staticmethod
    def inv_trsf(trsf: Union[np.ndarray, List[List]]) -> np.ndarray:
        """
        Inverse a given transformation

        Args:
            trsf (np.ndarray): 4x4 ndarray

        Returns:
            (np.ndarray): the 4x4 inverted matrix
        """
        return np.linalg.lstsq(trsf, np.identity(4))[0]

    def vox_to_real(
        self,
        trsf: Union[np.ndarray, List[List]],
        ref_vs: List[float],
    ) -> np.ndarray:
        """
        Transform a transformation for voxel units to physical units

        Args:
            trsf (ndarray): 4x4 matrix
            ref_vs (list(float, float, float)): initial voxel size in each dimension

        Returns:
            (np.ndarray): new matrix in metric size
        """
        H_ref = [
            [ref_vs[0], 0, 0, 0],
            [0, ref_vs[1], 0, 0],
            [0, 0, ref_vs[2], 0],
            [0, 0, 0, 1],
        ]
        H_ref_inv = self.inv_trsf(H_ref)
        return np.dot(trsf, H_ref_inv)

    def compute_trsfs(self, p: trsf_parameters):
        """
        Here is where the magic happens, give as an input the trsf_parameters object, get your transformations computed

        Args:
            p (trsf_parameters): parameters to compute the transformation
        """
        for A_num, flo_A in enumerate(p.flo_As):
            flo_voxel = p.flo_voxels[A_num]
            init_trsf = p.init_trsfs[A_num]
            trsf_path = p.trsf_paths[A_num]
            trsf_name = p.trsf_names[A_num]
            if isinstance(init_trsf, list):
                i = 0
                trsfs = []
                im_size = (
                    np.array(p.flo_im_sizes[A_num], dtype=float) * flo_voxel
                )
                while i < len(init_trsf):
                    t_type = init_trsf[i]
                    i += 1
                    axis = init_trsf[i]
                    i += 1
                    if "rot" in t_type:
                        angle = init_trsf[i]
                        i += 1
                        trsfs += [
                            self.axis_rotation_matrix(
                                axis.upper(), angle, np.zeros(3), im_size
                            )
                        ]
                    elif "flip" in t_type:
                        trsfs += [self.flip_matrix(axis.upper(), im_size)]
                    elif "trans" in t_type:
                        tr = init_trsf[i]
                        i += 1
                        trsfs += [self.translation_matrix(axis, tr)]
                res = np.identity(4)
                for trsf in trsfs:
                    res = np.dot(res, trsf)
                if not os.path.exists(trsf_path):
                    os.makedirs(trsf_path)
                init_trsf = os.path.join(
                    trsf_path, "A{:d}-init.trsf".format(A_num + 1)
                )
                np.savetxt(init_trsf, res)
            elif not p.init_trsf_real_unit and init_trsf is not None:
                tmp = self.vox_to_real(
                    self.inv_trsf(self.read_trsf(init_trsf)),
                    flo_voxel,
                    p.ref_voxel,
                )
                init_ext = init_trsf.split(".")[-1]
                init_trsf = init_trsf.replace(init_ext, "real.txt")
                np.savetxt(init_trsf, self.inv_trsf(tmp))
            if init_trsf is not None:
                init_trsf_command = " -init-trsf {:s}".format(init_trsf)
            else:
                init_trsf_command = ""
            i = 0
            if not p.test_init:
                for i, trsf_type in enumerate(p.trsf_types[:-1]):
                    if i != 0:
                        init_trsf_command = " -init-trsf {:s}".format(
                            os.path.join(trsf_path, res_trsf)
                        )
                    res_trsf = os.path.join(
                        trsf_path,
                        trsf_name.format(a=A_num + 1, trsf=trsf_type),
                    )
                    call(
                        p.path_to_bin
                        + "blockmatching -ref "
                        + p.ref_A.format(t=p.begin)
                        + " -flo "
                        + flo_A.format(t=p.begin)
                        + " -reference-voxel %f %f %f" % p.ref_voxel
                        + " -floating-voxel %f %f %f" % flo_voxel
                        + " -trsf-type %s -py-hl %d -py-ll %d"
                        % (
                            trsf_type,
                            p.registration_depth_start,
                            p.registration_depth_end,
                        )
                        + init_trsf_command
                        + " -res-trsf "
                        + res_trsf
                        + " -composition-with-initial",
                        shell=True,
                    )
                trsf_type = p.trsf_types[-1]
                i = len(p.trsf_types) - 1
                if i != 0:
                    init_trsf_command = " -init-trsf {:s}".format(
                        os.path.join(trsf_path, res_trsf)
                    )
                res_trsf = os.path.join(
                    trsf_path, trsf_name.format(a=A_num + 1, trsf=trsf_type)
                )
                res_inv_trsf = os.path.join(
                    trsf_path,
                    ("inv-" + trsf_name).format(a=A_num + 1, trsf=trsf_type),
                )
                call(
                    p.path_to_bin
                    + "blockmatching -ref "
                    + p.ref_A.format(t=p.begin)
                    + " -flo "
                    + flo_A.format(t=p.begin)
                    + " -reference-voxel %f %f %f" % p.ref_voxel
                    + " -floating-voxel %f %f %f" % flo_voxel
                    + " -trsf-type %s -py-hl %d -py-ll %d"
                    % (
                        trsf_type,
                        p.registration_depth_start,
                        p.registration_depth_end,
                    )
                    + init_trsf_command
                    + " -res-trsf "
                    + res_trsf
                    +  # ' -res-voxel-trsf ' + res_voxel_trsf + \
                    # ' -res ' + flo_out +\
                    " -composition-with-initial",
                    shell=True,
                )
                call(
                    p.path_to_bin + "invTrsf %s %s" % (res_trsf, res_inv_trsf),
                    shell=True,
                )

    @staticmethod
    def pad_trsfs(p: trsf_parameters, t: int = None):
        """
        Pad transformations

        Args:
            p (trsf_parameters): parameter object
            trsf_fmt (srt): the string format for the initial transformations
        """

        out_voxel = p.out_voxel
        trsf_path = p.trsf_paths[0]
        trsf_name = p.trsf_names[0]
        trsf_type = p.trsf_types[-1]

        im_shape = imread(p.ref_A.format(t=t)).shape
        im = SpatialImage(np.ones(im_shape, dtype=np.uint8))
        im.voxelsize = p.ref_voxel
        template = os.path.join(trsf_path, "tmp.tif")
        res_t = "template.tif"
        imsave(template, im)
        identity = np.identity(4)

        if p.test_init:
            trsf_name_only_a = "A{a:d}-init.trsf"
        else:
            where_a = trsf_name.find("{a:d}")
            no_a = (trsf_name[:where_a] + trsf_name[where_a + 5 :]).format(
                trsf=trsf_type
            )
            trsf_name_only_a = no_a[:where_a] + "{a:d}" + no_a[where_a:]

        trsf_fmt = os.path.join(trsf_path, trsf_name_only_a)
        trsf_fmt_no_flo = trsf_fmt.replace("{a:d}", "%d")
        new_trsf_fmt = ".".join(trsf_fmt.split(".")[:-1]) + "-padded.txt"
        new_trsf_fmt_no_flo = new_trsf_fmt.replace("{a:d}", "%d")
        np.savetxt(trsf_fmt.format(a=0, trsf=trsf_type), identity)

        call(
            p.path_to_bin
            + "changeMultipleTrsfs -trsf-format "
            + trsf_fmt_no_flo
            + " -index-reference %d -first %d -last %d "
            % (0, 0, len(p.trsf_paths))
            + " -template "
            + template
            + " -res "
            + new_trsf_fmt_no_flo
            + " -res-t "
            + os.path.join(trsf_path, res_t)
            + " "
            + " -trsf-type %s" % (trsf_type),
            shell=True,
        )
        template = os.path.join(trsf_path, res_t)
        return new_trsf_fmt, template

    def apply_trsf(self, p, t=None):
        """
        Apply the transformation according to `trsf_parameters`

        Args:
            p (trsf_parameters): parameters for the transformation
        """
        if p.bbox_out:
            trsf_fmt, template = self.pad_trsfs(p, t)
            if p.out_voxel != p.ref_voxel:
                before, after = os.path.splitext(template)
                old_template = template
                template = "".join((before, ".final_template", after))
                call(
                    p.path_to_bin
                    + f"applyTrsf {old_template} {template} "
                    + "-vs %f %f %f" % p.out_voxel,
                    shell=True,
                )
            A0_trsf = (
                " -trsf " + trsf_fmt.format(a=0) + " -template " + template
            )
        else:
            A0_trsf = ""
        if p.out_voxel != p.ref_voxel or p.bbox_out:
            call(
                p.path_to_bin
                + "applyTrsf"
                + " -flo "
                + p.ref_A.format(t=t)
                + " -res "
                + p.ref_out.format(t=t)
                + " -floating-voxel %f %f %f" % p.ref_voxel
                + " -vs %f %f %f" % p.out_voxel
                + A0_trsf,
                shell=True,
            )
        elif p.copy_ref:
            copyfile(p.ref_A.format(t=t), p.ref_out.format(t=t))
        else:
            p.ref_out = p.ref_A
        for A_num, flo_A in enumerate(p.flo_As):
            flo_voxel = p.flo_voxels[A_num]
            trsf_path = p.trsf_paths[A_num]
            trsf_name = p.trsf_names[A_num]
            init_trsf = p.init_trsfs[A_num]
            if p.test_init:
                if isinstance(init_trsf, list):
                    trsf = " -trsf " + os.path.join(
                        trsf_path, "A{:d}-init.trsf".format(A_num + 1)
                    )
                else:
                    trsf = " -trsf " + init_trsf
            elif not p.bbox_out:
                t_type = "" if len(p.trsf_types) < 1 else p.trsf_types[-1]
                trsf = " -trsf " + os.path.join(
                    trsf_path, trsf_name.format(a=A_num + 1, trsf=t_type)
                )
            else:
                trsf = (
                    " -trsf "
                    + trsf_fmt.format(a=A_num + 1)
                    + " -template "
                    + template
                )
            flo_out = p.flo_outs[A_num]
            call(
                p.path_to_bin
                + "applyTrsf"
                + " -flo "
                + flo_A.format(t=t)
                + " -floating-voxel %f %f %f" % flo_voxel
                + " -res "
                + flo_out.format(t=t)
                + " -ref "
                + p.ref_out.format(t=t)
                + " -reference-voxel %f %f %f" % p.out_voxel
                + trsf
                + " -interpolation %s" % p.image_interpolation,
                shell=True,
            )

    @staticmethod
    def prettify(elem: ET.Element) -> str:
        """
        Return a pretty-printed XML string for the Element.

        Args:
            elem (xml.etree.ElementTree.Element): xml element

        Returns:
            (str): a nice version of our xml file
        """
        rough_string = ET.tostring(elem, "utf-8")
        reparsed = minidom.parseString(rough_string)
        return reparsed.toprettyxml(indent="  ")

    @staticmethod
    def do_viewSetup(
        ViewSetup: ET.SubElement,
        p: trsf_parameters,
        im_size: Tuple[int, int, int],
        i: int,
    ):
        """
        Setup xml elements for BigDataViewer

        Args:
            ViewSetup (ET.SubElement): ...
            p (trsf_parameters): Parameter object
            im_size (tuple): tuple of 3 integers with the dimension of the image
            i (int): angle id
        """
        id_ = ET.SubElement(ViewSetup, "id")
        id_.text = "%d" % i
        name = ET.SubElement(ViewSetup, "name")
        name.text = "%d" % i
        size = ET.SubElement(ViewSetup, "size")
        size.text = "%d %d %d" % tuple(im_size)
        voxelSize = ET.SubElement(ViewSetup, "voxelSize")
        unit = ET.SubElement(voxelSize, "unit")
        unit.text = p.bdv_unit
        size = ET.SubElement(voxelSize, "size")
        size.text = "%f %f %f" % tuple(p.bdv_voxel_size)
        attributes = ET.SubElement(ViewSetup, "attributes")
        illumination = ET.SubElement(attributes, "illumination")
        illumination.text = "0"
        channel = ET.SubElement(attributes, "channel")
        channel.text = "0"
        tile = ET.SubElement(attributes, "tile")
        tile.text = "0"
        angle = ET.SubElement(attributes, "angle")
        angle.text = "%d" % i

    def do_ViewRegistration(
        self,
        ViewRegistrations: ET.SubElement,
        p: trsf_parameters,
        t: int,
        a: int,
    ):
        """
        Write the view registration for BigDataViewer

        Args:
            ViewSetup (ET.SubElement): ...
            p (trsf_parameters): Parameter object
            t (int): time point to treat
            a (int): angle to treat
        """

        ViewRegistration = ET.SubElement(ViewRegistrations, "ViewRegistration")
        ViewRegistration.set("timepoint", "%d" % t)
        ViewRegistration.set("setup", "%d" % a)
        ViewTransform = ET.SubElement(ViewRegistration, "ViewTransform")
        ViewTransform.set("type", "affine")
        if a != 0:
            affine = ET.SubElement(ViewTransform, "affine")
            trsf_path = p.trsf_paths[a - 1]
            trsf_name = p.trsf_names[a - 1]
            trsf_type = p.trsf_types[-1]
            f = os.path.join(
                trsf_path, ("inv-" + trsf_name).format(a=a, trsf=trsf_type)
            )
            trsf = self.read_trsf(f)
            formated_trsf = tuple(trsf[:-1, :].flatten())
            affine.text = ("%f " * 12) % formated_trsf
            ViewTransform = ET.SubElement(ViewRegistration, "ViewTransform")
            ViewTransform.set("type", "affine")
            affine = ET.SubElement(ViewTransform, "affine")
            affine.text = (
                "%f 0.0 0.0 0.0 0.0 %f 0.0 0.0 0.0 0.0 %f 0.0"
                % p.flo_voxels[a - 1]
            )
        else:
            affine = ET.SubElement(ViewTransform, "affine")
            affine.text = (
                "%f 0.0 0.0 0.0 0.0 %f 0.0 0.0 0.0 0.0 %f 0.0" % p.ref_voxel
            )

    def build_bdv(self, p: trsf_parameters):
        """
        Build the BigDataViewer xml

        Args:
            p (trsf_parameters): Parameter object
        """
        SpimData = ET.Element("SpimData")
        SpimData.set("version", "0.2")

        base_path = ET.SubElement(SpimData, "BasePath")
        base_path.set("type", "relative")
        base_path.text = "."

        SequenceDescription = ET.SubElement(SpimData, "SequenceDescription")

        ImageLoader = ET.SubElement(SequenceDescription, "ImageLoader")
        ImageLoader.set("format", "klb")
        Resolver = ET.SubElement(ImageLoader, "Resolver")
        Resolver.set(
            "type", "org.janelia.simview.klb.bdv.KlbPartitionResolver"
        )
        for im_p in p.bdv_im:
            ViewSetupTemplate = ET.SubElement(Resolver, "ViewSetupTemplate")
            template = ET.SubElement(ViewSetupTemplate, "template")
            template.text = im_p
            timeTag = ET.SubElement(ViewSetupTemplate, "timeTag")
            timeTag.text = p.time_tag

        ViewSetups = ET.SubElement(SequenceDescription, "ViewSetups")
        ViewSetup = ET.SubElement(ViewSetups, "ViewSetup")
        i = 0
        self.do_viewSetup(ViewSetup, p, p.ref_im_size, i)
        for i, pi in enumerate(p.flo_voxels):
            ViewSetup = ET.SubElement(ViewSetups, "ViewSetup")
            self.do_viewSetup(ViewSetup, p, p.flo_im_sizes[i], i + 1)

        Attributes = ET.SubElement(ViewSetups, "Attributes")
        Attributes.set("name", "illumination")
        Illumination = ET.SubElement(Attributes, "Illumination")
        id_ = ET.SubElement(Illumination, "id")
        id_.text = "0"
        name = ET.SubElement(Illumination, "name")
        name.text = "0"

        Attributes = ET.SubElement(ViewSetups, "Attributes")
        Attributes.set("name", "channel")
        Channel = ET.SubElement(Attributes, "Channel")
        id_ = ET.SubElement(Channel, "id")
        id_.text = "0"
        name = ET.SubElement(Channel, "name")
        name.text = "0"

        Attributes = ET.SubElement(ViewSetups, "Attributes")
        Attributes.set("name", "tile")
        Tile = ET.SubElement(Attributes, "Tile")
        id_ = ET.SubElement(Tile, "id")
        id_.text = "0"
        name = ET.SubElement(Tile, "name")
        name.text = "0"

        Attributes = ET.SubElement(ViewSetups, "Attributes")
        Attributes.set("name", "angle")
        for i in range(len(p.flo_voxels) + 1):
            Angle = ET.SubElement(Attributes, "Angle")
            id_ = ET.SubElement(Angle, "id")
            id_.text = "%d" % i
            name = ET.SubElement(Angle, "name")
            name.text = "%d" % i

        TimePoints = ET.SubElement(SequenceDescription, "Timepoints")
        TimePoints.set("type", "range")
        first = ET.SubElement(TimePoints, "first")
        first.text = "%d" % p.begin
        last = ET.SubElement(TimePoints, "last")
        last.text = "%d" % p.end
        ViewRegistrations = ET.SubElement(SpimData, "ViewRegistrations")
        for t in range(p.begin, p.end + 1):
            self.do_ViewRegistration(ViewRegistrations, p, t, 0)
            for a in range(len(p.flo_voxels)):
                self.do_ViewRegistration(ViewRegistrations, p, t, a + 1)

        with open(p.out_bdv, "w") as f:
            f.write(self.prettify(SpimData))
            f.close()

    def run_trsf(self):
        """
        Start the Spatial registration after having informed the parameter files
        """
        for p in self.params:
            try:
                print("Starting experiment")
                print(p)
                self.prepare_paths(p)
                if p.compute_trsf or p.test_init:
                    self.compute_trsfs(p)
                if p.apply_trsf or p.test_init:
                    if not (p.begin is None and p.end is None):
                        for t in range(p.begin, p.end + 1):
                            self.apply_trsf(p, t)
                    else:
                        self.apply_trsf(p)
                if p.do_bdv:
                    self.build_bdv(p)
            except Exception as e:
                print("Failure of %s" % p.origin_file_name)
                print(e)

    def __init__(self, params=None):
        if params is None:
            self.params = self.read_param_file()
        elif (
            isinstance(params, str)
            or isinstance(params, Path)
            or isinstance(params, dict)
        ):
            self.params = SpatialRegistration.read_param_file(params)
        else:
            self.params = params
        if self.params is not None and 0 < len(self.params):
            self.path_to_bin = self.params[0].path_to_bin


def spatial_registration():
    reg = SpatialRegistration()
    reg.run_trsf()
